# SPDX-License-Identifier: MIT
# Copyright (c) 2015, Elaborated Networks GmbH
# Copyright (c) 2019, Mathias Laurin

"""Ciphers for symmetric encryption and decryption."""


cimport mbedtls.cipher._cipher as _cipher
from libc.stdlib cimport malloc, free

import enum
from contextlib import suppress

import mbedtls.exceptions as _exc


CIPHER_NAME = (
    # Define as bytes to map to `const char*` without conversion.
    b"NONE",
    b"NULL",
    b"AES-128-ECB",
    b"AES-192-ECB",
    b"AES-256-ECB",
    b"AES-128-CBC",
    b"AES-192-CBC",
    b"AES-256-CBC",
    b"AES-128-CFB128",
    b"AES-192-CFB128",
    b"AES-256-CFB128",
    b"AES-128-CTR" ,
    b"AES-192-CTR",
    b"AES-256-CTR",
    b"AES-128-GCM",
    b"AES-192-GCM",
    b"AES-256-GCM",
    b"CAMELLIA-128-ECB",
    b"CAMELLIA-192-ECB",
    b"CAMELLIA-256-ECB",
    b"CAMELLIA-128-CBC",
    b"CAMELLIA-192-CBC",
    b"CAMELLIA-256-CBC",
    b"CAMELLIA-128-CFB128",
    b"CAMELLIA-192-CFB128",
    b"CAMELLIA-256-CFB128",
    b"CAMELLIA-128-CTR",
    b"CAMELLIA-192-CTR",
    b"CAMELLIA-256-CTR",
    b"CAMELLIA-128-GCM",
    b"CAMELLIA-192-GCM",
    b"CAMELLIA-256-GCM",
    b"DES-ECB",
    b"DES-CBC",
    b"DES-EDE-ECB",
    b"DES-EDE-CBC",
    b"DES-EDE3-ECB",
    b"DES-EDE3-CBC",
    b"BLOWFISH-ECB",
    b"BLOWFISH-CBC",
    b"BLOWFISH-CFB64",
    b"BLOWFISH-CTR",
    b"ARC4-128",
    b"AES-128-CCM",
    b"AES-192-CCM",
    b"AES-256-CCM",
    b"CAMELLIA-128-CCM",
    b"CAMELLIA-192-CCM",
    b"CAMELLIA-256-CCM",
    b"ARIA-128-ECB",
    b"ARIA-192-ECB",
    b"ARIA-256-ECB",
    b"ARIA-128-CBC",
    b"ARIA-192-CBC",
    b"ARIA-256-CBC",
    b"ARIA-128-CFB128",
    b"ARIA-192-CFB128",
    b"ARIA-256-CFB128",
    b"ARIA-128-CTR",
    b"ARIA-192-CTR",
    b"ARIA-256-CTR",
    b"ARIA-128-GCM",
    b"ARIA-192-GCM",
    b"ARIA-256-GCM",
    b"ARIA-128-CCM",
    b"ARIA-192-CCM",
    b"ARIA-256-CCM",
    b"AES-128-OFB",
    b"AES-192-OFB",
    b"AES-256-OFB",
    b"AES-128-XTS",
    b"AES-256-XTS",
    b"CHACHA20",
    b"CHACHA20-POLY1305",
)


@enum.unique
class Mode(enum.Enum):
    ECB = _cipher.MBEDTLS_MODE_ECB
    CBC = _cipher.MBEDTLS_MODE_CBC
    CFB = _cipher.MBEDTLS_MODE_CFB
    OFB = _cipher.MBEDTLS_MODE_OFB
    CTR = _cipher.MBEDTLS_MODE_CTR
    GCM = _cipher.MBEDTLS_MODE_GCM
    STREAM = _cipher.MBEDTLS_MODE_STREAM
    CCM = _cipher.MBEDTLS_MODE_CCM
    XTS = _cipher.MBEDTLS_MODE_XTS
    CHACHAPOLY = _cipher.MBEDTLS_MODE_CHACHAPOLY


# Add module-level aliases to comply with PEP 272.
MODE_ECB = Mode.ECB
MODE_CBC = Mode.CBC
MODE_CFB = Mode.CFB
MODE_OFB = Mode.OFB
MODE_CTR = Mode.CTR
MODE_GCM = Mode.GCM
MODE_STREAM = Mode.STREAM
MODE_CCM = Mode.CCM
MODE_XTS = Mode.XTS
MODE_CHACHAPOLY = Mode.CHACHAPOLY


__all__ = (
    "Mode",
    "MODE_ECB",
    "MODE_CBC",
    "MODE_CFB",
    "MODE_OFB",
    "MODE_CTR",
    "MODE_GCM",
    "MODE_STREAM",
    "MODE_CCM",
    "MODE_XTS",
    "MODE_CHACHAPOLY",
)


cpdef get_supported_ciphers():
    """Return the set of ciphers supported by the generic
    cipher module.

    """
    cipher_lookup = {n: v for n, v in enumerate(CIPHER_NAME)}
    cdef const int* cipher_types = _cipher.mbedtls_cipher_list()
    cdef size_t n = 0
    ciphers = set()
    while cipher_types[n]:
        with suppress(KeyError):
            ciphers.add(cipher_lookup[cipher_types[n]])
        n += 1
    return ciphers


cdef class _CipherBase:

    """Wrap and encapsulate the cipher library from mbed TLS.

    Parameters:
        name (bytes): The cipher name known to mbed TLS.
        key (bytes or None): The key to encrypt decrypt.  If None,
            encryption and decryption are unavailable.
        mode (int): The mode of operation of the cipher.
        iv (bytes): The initialization vector (IV).  The IV is
            required for every mode but ECB and CTR where it is ignored.

        Attributes:
            block_size (int): The block size for the cipher in bytes.
            mode (int): The mode of operation of the cipher.
            iv_size (int): The size of the cipher's IV/NONCE in bytes.
            key_size (int): The size of the cipher's key, in bytes.

    """
    def __init__(
        self,
        cipher_name,
        const unsigned char[:] key,
        mode,
        const unsigned char[:] iv not None
    ):
        mode = Mode(mode)
        if mode in {Mode.CBC, Mode.CFB} and iv.size == 0:
            raise ValueError("mode requires an IV")
        if cipher_name not in get_supported_ciphers():
            raise NotImplementedError("unsupported cipher: %r" % cipher_name)

        _exc.check_error(_cipher.mbedtls_cipher_setup(
            &self._enc_ctx,
            _cipher.mbedtls_cipher_info_from_string(cipher_name)))
        _exc.check_error(_cipher.mbedtls_cipher_setup(
            &self._dec_ctx,
            _cipher.mbedtls_cipher_info_from_string(cipher_name)))
        # Remove the default padding for the modes that support it and
        # ignore possible errors caused by a cipher mode that doesn't.
        #
        # Note: Padding is only supported by CBC (mbedtls 2.16.12).
        _cipher.mbedtls_cipher_set_padding_mode(
            &self._enc_ctx, _cipher.MBEDTLS_PADDING_NONE)
        _cipher.mbedtls_cipher_set_padding_mode(
            &self._dec_ctx, _cipher.MBEDTLS_PADDING_NONE)

        if key is not None:
            _exc.check_error(_cipher.mbedtls_cipher_setkey(
                &self._enc_ctx, &key[0], 8 * key.size,
                _cipher.MBEDTLS_ENCRYPT))
            _exc.check_error(_cipher.mbedtls_cipher_setkey(
                &self._dec_ctx, &key[0], 8 * key.size,
                _cipher.MBEDTLS_DECRYPT))

        _exc.check_error(_cipher.mbedtls_cipher_set_iv(
            &self._enc_ctx, &iv[0] if iv.size else NULL, iv.size))
        _exc.check_error(_cipher.mbedtls_cipher_set_iv(
            &self._dec_ctx, &iv[0] if iv.size else NULL, iv.size))

    def __cinit__(self):
        """Initialize a `cipher_context` (as NONE)."""
        _cipher.mbedtls_cipher_init(&self._enc_ctx)
        _cipher.mbedtls_cipher_init(&self._dec_ctx)

    def __dealloc__(self):
        """Free and clear the cipher-specific context of ctx."""
        _cipher.mbedtls_cipher_free(&self._enc_ctx)
        _cipher.mbedtls_cipher_free(&self._dec_ctx)

    def __getstate__(self):
        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")

    def __str__(self):
        """Return the name of the cipher."""
        return self.name.decode("ascii")

    @property
    def block_size(self):
        """Return the block size for the cipher."""
        return _cipher.mbedtls_cipher_get_block_size(&self._enc_ctx)

    @property
    def mode(self):
        """Return the mode of operation of the cipher."""
        return Mode(_cipher.mbedtls_cipher_get_cipher_mode(&self._enc_ctx))

    @property
    def iv_size(self):
        """Return the size of the cipher's IV/NONCE in bytes."""
        return _cipher.mbedtls_cipher_get_iv_size(&self._enc_ctx)

    @property
    def _type(self):
        """Return the type of the cipher."""
        return _cipher.mbedtls_cipher_get_type(&self._enc_ctx)

    @property
    def name(self):
        """Return the name of the cipher."""
        ret = _cipher.mbedtls_cipher_get_name(&self._enc_ctx)
        return ret if ret is not NULL else b"NONE"

    @property
    def key_size(self):
        """Return the size of the ciphers' key."""
        return _cipher.mbedtls_cipher_get_key_bitlen(&self._enc_ctx) // 8

    def encrypt(self, const unsigned char[:] message not None):
        raise NotImplementedError

    def decrypt(self, const unsigned char[:] message not None):
        raise NotImplementedError


cdef class Cipher(_CipherBase):
    cdef _crypt(self,
                _cipher.mbedtls_cipher_context_t *ctx,
                const unsigned char[:] input):
        if input.size == 0:
            _exc.check_error(-0x6280)  # Raise full block expected error.
        cdef size_t olen
        cdef size_t finish_olen
        cdef size_t sz = input.size + self.block_size
        cdef unsigned char* output = <unsigned char*>malloc(
            sz * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(_cipher.mbedtls_cipher_reset(ctx))
            _exc.check_error(_cipher.mbedtls_cipher_update(
                ctx, &input[0], input.size, output, &olen))
            err = _cipher.mbedtls_cipher_finish(ctx, output + olen, &finish_olen)
            if err == -0x6280:
                raise ValueError("expected a full block")
            _exc.check_error(err)
            return output[:olen + finish_olen]
        finally:
            free(output)

    def encrypt(self, const unsigned char[:] message not None):
        return self._crypt(&self._enc_ctx, message)

    def decrypt(self, const unsigned char[:] message not None):
        return self._crypt(&self._dec_ctx, message)


cdef class AEADCipher(_CipherBase):
    def __init__(self,
                 cipher_name,
                 const unsigned char[:] key,
                 mode,
                 const unsigned char[:] iv not None,
                 const unsigned char[:] ad not None):
        super().__init__(cipher_name, key, mode, iv)
        self._iv = iv
        self._ad = ad

    cdef _aead_encrypt(self,
                const unsigned char[:] iv,
                const unsigned char[:] ad,
                const unsigned char[:] input):
        if input.size == 0:
            _exc.check_error(-0x6280)  # Raise full block expected error.
        assert iv.size != 0
        cdef size_t olen
        cdef size_t sz = input.size + self.block_size
        cdef unsigned char tag[16];
        cdef unsigned char* output = <unsigned char*>malloc(
            sz * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            if ad.size:
                pad = <const unsigned char*> &ad[0]
            else:
                pad = NULL
            _exc.check_error(_cipher.mbedtls_cipher_auth_encrypt(
                &self._enc_ctx,
                &iv[0], iv.size, pad, ad.size,
                &input[0], input.size, output, &olen,
                tag, sizeof(tag)))
            return output[:olen], tag[:16]
        finally:
            free(output)

    cdef _aead_decrypt(self,
                const unsigned char[:] iv,
                const unsigned char[:] ad,
                const unsigned char[:] input,
                const unsigned char[:] tag):
        if input.size == 0:
            _exc.check_error(-0x6280)  # Raise full block expected error.
        assert iv.size != 0
        assert tag.size == 16

        cdef const unsigned char *pad
        if ad.size == 0:
            pad = NULL
        else:
            pad = &ad[0]

        cdef size_t olen
        cdef size_t sz = input.size + self.block_size
        cdef unsigned char* output = <unsigned char*>malloc(
            sz * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(_cipher.mbedtls_cipher_auth_decrypt(
                &self._dec_ctx,
                &iv[0], iv.size, pad, ad.size,
                &input[0], input.size, output, &olen,
                &tag[0], tag.size))
            return output[:olen]
        finally:
            free(output)

    def encrypt(self, const unsigned char[:] message not None):
        return self._aead_encrypt(self._iv, self._ad, message)

    def decrypt(self, const unsigned char[:] message not None,
                const unsigned char[:] tag not None):
        return self._aead_decrypt(self._iv, self._ad, message, tag)



# SPDX-License-Identifier: MIT
# Copyright (c) 2018, Mathias Laurin

import enum
import socket as _socket
import struct

from ._tls import (
    DTLSConfiguration,
    DTLSVersion,
    HandshakeStep,
    HelloVerifyRequest,
)
from ._tls import MbedTLSBuffer as TLSWrappedBuffer
from ._tls import (
    NextProtocol,
    Purpose,
    RaggedEOF,
    TLSConfiguration,
    TLSSession,
    TLSVersion,
    TrustStore,
    WantReadError,
    WantWriteError,
    _BaseContext,
    ciphers_available,
)

__all__ = (
    "ClientContext",
    "DTLSConfiguration",
    "DTLSVersion",
    "HelloVerifyRequest",
    "NextProtocol",
    "Purpose",
    "RaggedEOF",
    "ServerContext",
    "TLSConfiguration",
    "TLSRecordHeader",
    "TLSSession",
    "TLSVersion",
    "TLSWrappedBuffer",
    "TLSWrappedSocket",
    "TrustStore",
    "WantReadError",
    "WantWriteError",
    "ciphers_available",
)


class TLSRecordHeader:
    """Encode/decode TLS record protocol format."""

    __slots__ = ("record_type", "version", "length")
    fmt = "!BHH"

    class RecordType(enum.IntEnum):
        CHANGE_CIPHER_SPEC = 0x14
        ALERT = 0x15
        HANDSHAKE = 0x16
        APPLICATION_DATA = 0x17

    def __init__(self, record_type, version, length):
        self.record_type = record_type
        self.version = version
        self.length = length

    def __str__(self):
        return "%s(%s, %s, %s)" % (
            type(self).__name__,
            self.record_type,
            self.version,
            self.length,
        )

    def __repr__(self):
        return "%s(%r, %r, %r)" % (
            type(self).__name__,
            self.record_type,
            self.version,
            self.length,
        )

    def __eq__(self, other):
        if not isinstance(other, TLSRecordHeader):
            return NotImplemented
        return (
            self.record_type is other.record_type
            and self.version is other.version
            and self.length == other.length
        )

    def __hash__(self):
        return 0x5AFE ^ self.record_type ^ self.version ^ self.length

    def __len__(self):
        return 5

    def __bytes__(self):
        return struct.pack(
            TLSRecordHeader.fmt, self.record_type, self.version, self.length
        )

    @classmethod
    def from_bytes(cls, header):
        record_type, version, length = struct.unpack(
            TLSRecordHeader.fmt, header[:5]
        )
        return cls(
            TLSRecordHeader.RecordType(record_type),
            TLSVersion(version),
            length,
        )


class ClientContext(_BaseContext):
    # _pep543.ClientContext

    @property
    def _purpose(self) -> Purpose:
        return Purpose.CLIENT_AUTH

    def wrap_socket(self, socket, server_hostname):
        """Wrap an existing Python socket object ``socket`` and return a
        ``TLSWrappedSocket`` object. ``socket`` must be a ``SOCK_STREAM``
        socket: all other socket types are unsupported.

        Args:
            socket (socket.socket): The socket to wrap.
            server_hostname (str, optional): The hostname of the service
                which we are connecting to.  Pass ``None`` if hostname
                validation is not desired.  This parameter has no
                default value because opting-out hostname validation is
                dangerous and should not be the default behavior.

        """
        buffer = self.wrap_buffers(server_hostname)
        return TLSWrappedSocket(socket, buffer)

    def wrap_buffers(self, server_hostname):
        """Create an in-memory stream for TLS."""
        # PEP 543
        return TLSWrappedBuffer(self, server_hostname)


class ServerContext(_BaseContext):
    # _pep543.ServerContext

    @property
    def _purpose(self) -> Purpose:
        return Purpose.SERVER_AUTH

    def wrap_socket(self, socket):
        """Wrap an existing Python socket object ``socket``."""
        buffer = self.wrap_buffers()
        return TLSWrappedSocket(socket, buffer)

    def wrap_buffers(self):
        # PEP 543
        return TLSWrappedBuffer(self)


class TLSWrappedSocket:
    # _pep543.TLSWrappedSocket
    def __init__(self, socket, buffer):
        super().__init__()
        self._socket = socket
        self._buffer = buffer
        self._context = buffer.context
        self._closed = False

    def __getstate__(self):
        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        if not self._closed:
            self.close()

    def __str__(self):
        return str(self._socket)

    @property
    def _handshake_state(self):
        return self._buffer._handshake_state

    # PEP 543 requires the full socket API.

    @property
    def family(self):
        return self._socket.family

    @property
    def proto(self):
        return self._socket.proto

    @property
    def type(self):
        return self._socket.type

    def accept(self):
        if self.type == _socket.SOCK_STREAM:
            conn, address = self._socket.accept()
        else:
            _, address = self._socket.recvfrom(1024, _socket.MSG_PEEK)
            # Use this socket to communicate with the client and bind
            # another one for the next connection.  This procedure is
            # adapted from `mbedtls_net_accept()`.
            sockname = self.getsockname()
            conn = _socket.fromfd(self.fileno(), self.family, self.type)
            conn.connect(address)
            # Closing the socket on Python 2.7 and 3.4 invalidates
            # the accessors.  So we should get the values first.
            family, type_, proto = self.family, self.type, self.proto
            self.close()
            self._socket = _socket.socket(family, type_, proto)
            self.setsockopt(_socket.SOL_SOCKET, _socket.SO_REUSEADDR, 1)
            self.bind(sockname)
        return self.context.wrap_socket(conn), address

    def bind(self, address):
        self._socket.bind(address)

    def close(self):
        self._closed = True
        self._buffer.shutdown()
        self._socket.close()

    def connect(self, address):
        self._socket.connect(address)

    def connect_ex(self, address):
        self._socket.connect_ex(address)

    def fileno(self):
        return self._socket.fileno()

    def getpeername(self):
        return self._socket.getpeername()

    def getsockname(self):
        return self._socket.getsockname()

    def getsockopt(self, optname, buflen=None):
        return self._socket.getsockopt(optname, buflen=buflen)

    def listen(self, backlog=None):
        if backlog is None:
            # Use 5 (Python default) or 10 (mbedtls defaults).
            backlog = 5
        self._socket.listen(backlog)

    def makefile(self, *args, **kwargs):
        return self._socket.makefile(*args, **kwargs)

    def recv(self, bufsize, flags=0):
        encrypted = self._socket.recv(bufsize, flags)
        if not encrypted:
            return b""
        self._buffer.receive_from_network(encrypted)
        return self._buffer.read(bufsize)

    def recv_into(self, buffer, nbytes=None, flags=0):
        raise NotImplementedError

    def recvfrom(self, bufsize, flags=0):
        encrypted, addr = self._socket.recvfrom(bufsize, flags)
        if not encrypted:
            return b"", addr
        self._buffer.receive_from_network(encrypted)
        return self._buffer.read(bufsize), addr

    def recvfrom_into(self, buffer, nbytes=None, flags=0):
        encrypted, addr = self._socket.recvfrom(len(buffer), flags)
        if not encrypted:
            return buffer, addr
        self._buffer.receive_from_network(encrypted)
        return self._buffer.readinto(buffer, nbytes), addr

    def send(self, message, flags=0):
        # Maximum size supported by TLS is 16K (encrypted).
        # mbedTLS defines it in MBEDTLS_SSL_MAX_CONTENT_LEN and
        # MBEDTLS_SSL_IN_CONTENT_LEN/MBEDTLS_SSL_OUT_CONTENT_LEN.
        amt = self._buffer.write(message)
        encrypted = self._buffer.peek_outgoing(amt)
        self._socket.send(encrypted, flags)
        self._buffer.consume_outgoing(amt)
        return len(message)

    def sendall(self, message, flags=0):
        amt = self._buffer.write(message)
        encrypted = self._buffer.peek_outgoing(amt)
        self._buffer.consume_outgoing(amt)
        self._socket.sendall(encrypted)

    def sendto(self, message, *args):
        if not 1 <= len(args) <= 2:
            raise TypeError(
                "sendto() takes 2 or 3 arguments (%i given)" % (1 + len(args))
            )
        if len(args) == 1:
            flags, address = 0, args[0]
        else:
            flags, address = args

        amt = self._buffer.write(message)
        encrypted = self._buffer.peek_outgoing(amt)
        if flags:
            self._socket.sendto(encrypted, flags, address)
        else:
            self._socket.sendto(encrypted, address)
        self._buffer.consume_outgoing(amt)
        return len(message)

    def setblocking(self, flag):
        self._socket.setblocking(flag)

    def settimeout(self, value):
        self._socket.settimeout(value)

    def gettimeout(self):
        return self._socket.gettimeout()

    def setsockopt(self, level, optname, value):
        self._socket.setsockopt(level, optname, value)

    def shutdown(self, how):
        self._buffer.shutdown()
        self._socket.shutdown(how)

    # PEP 543 adds the following methods.

    def do_handshake(self):
        while self._handshake_state is not HandshakeStep.HANDSHAKE_OVER:
            try:
                self._buffer.do_handshake()
            except WantReadError:
                data = self._socket.recv(1024)
                self._buffer.receive_from_network(data)
            except WantWriteError:
                in_transit = self._buffer.peek_outgoing(1024)
                amt = self._socket.send(in_transit)
                self._buffer.consume_outgoing(amt)

    def setcookieparam(self, param):
        self._buffer.setcookieparam(param)

    def cipher(self):
        return self._buffer.cipher()

    def negotiated_protocol(self):
        return self._buffer.negotiated_protocol()

    @property
    def context(self):
        return self._buffer.context

    def negotiated_tls_version(self):
        return self._buffer.negotiated_tls_version()

    def unwrap(self):
        self._buffer.shutdown()
        return self._socket



# SPDX-License-Identifier: MIT
# Copyright (c) 2016, Elaborated Networks GmbH
# Copyright (c) 2019, Mathias Laurin

"""Blowfish cipher designed by Bruce Schneier in 1993."""


from mbedtls.exceptions import TLSError

from . import _cipher

__all__ = ["block_size", "key_size", "new"]

block_size = 8
key_size = None


def new(key, mode, iv=None):
    """Return a `Cipher` object that can perform Blowfish encryption and
    decryption.

    Blowfish cipher designed by Bruce Schneier in 1993.

    Parameters:
        key (bytes or None): The key to encrypt decrypt.  If None,
            encryption and decryption are unavailable.
        mode (Mode): The mode of operation of the cipher.
        iv (bytes or None): The initialization vector (IV).  The IV is
            required for every mode but ECB and CTR where it is ignored.
            If not set, the IV is initialized to all 0, which should not
            be used for encryption.

    """
    mode = _cipher.Mode(mode)
    key_len = len(key)
    if key_len not in range(4, 57):
        raise TLSError(msg="key size must be 4 to 56 bytes, got %i" % key_len)
    if mode not in {
        _cipher.Mode.ECB,
        _cipher.Mode.CBC,
        _cipher.Mode.CFB,
        _cipher.Mode.CTR,
    }:
        raise TLSError(msg="unsupported mode %r" % mode)
    name = (
        "BLOWFISH-%s%s" % (mode.name, "64" if mode is _cipher.Mode.CFB else "")
    ).encode("ascii")
    return _cipher.Cipher(name, key, mode, iv)



"""Unit tests for mbedtls.md."""

# Disable checks for violations that are acceptable in tests.
# pylint: disable=missing-docstring
# pylint: disable=attribute-defined-outside-init
# pylint: disable=invalid-name

import pickle
from collections.abc import Collection

import pytest

import mbedtls
from mbedtls import hashlib
from mbedtls import hmac as hmaclib
from mbedtls._md import Hash


def make_chunks(buffer, size):
    for i in range(0, len(buffer), size):
        yield buffer[i : i + size]


def test_make_chunks(randbytes):
    buffer = randbytes(1024)
    assert b"".join(buf for buf in make_chunks(buffer, 100)) == buffer


def test_algorithms():
    assert isinstance(hashlib.algorithms_guaranteed, Collection)
    assert hashlib.algorithms_guaranteed

    assert isinstance(hashlib.algorithms_available, Collection)
    assert hashlib.algorithms_available

    assert frozenset(hashlib.algorithms_guaranteed).issubset(
        hashlib.algorithms_available
    )


def test_invalid_name_raises_TypeError():
    with pytest.raises(TypeError):
        Hash(42)


def test_unavailable_cipher_raises_ValueError():
    with pytest.raises(ValueError):
        Hash("unavailable")


class _TestMDBase:
    @pytest.fixture
    def algorithm(self):
        raise NotImplementedError

    @pytest.fixture
    def digest_size(self):
        raise NotImplementedError

    @pytest.fixture
    def block_size(self):
        raise NotImplementedError

    @pytest.mark.parametrize("repr_", (repr, str), ids=lambda f: f.__name__)
    def test_repr(self, repr_, algorithm):
        assert isinstance(repr_(algorithm), str)

    def test_pickle(self, algorithm):
        with pytest.raises(TypeError) as excinfo:
            pickle.dumps(algorithm)

        assert str(excinfo.value).startswith("cannot pickle")

    def test_digest_size_accessor(self, algorithm, digest_size):
        assert algorithm.digest_size == digest_size

    def test_digest_size(self, algorithm, digest_size):
        assert len(algorithm.digest()) == digest_size

    def test_block_size_accessor(self, algorithm, block_size):
        assert algorithm.block_size == block_size


class _TestHash(_TestMDBase):
    @pytest.fixture
    def buffer(self, randbytes):
        return randbytes(512)

    def test_new(self, algorithm, buffer):
        copy = hashlib.new(algorithm.name, buffer)
        algorithm.update(buffer)
        assert algorithm.digest() == copy.digest()
        assert algorithm.hexdigest() == copy.hexdigest()

    def test_copy_and_update(self, algorithm, buffer):
        copy = algorithm.copy()
        algorithm.update(buffer)
        copy.update(buffer)
        assert algorithm.digest() == copy.digest()
        assert algorithm.hexdigest() == copy.hexdigest()

    def test_copy_and_update_nothing(self, algorithm, buffer):
        copy = algorithm.copy()
        algorithm.update(b"")
        assert algorithm.digest() == copy.digest()
        assert algorithm.hexdigest() == copy.hexdigest()


@pytest.mark.skipif(
    not mbedtls.has_feature("md2"), reason="requires MD2 support in libmbedtls"
)
class TestHashMD2(_TestHash):
    @pytest.fixture
    def algorithm(self):
        return hashlib.md2()

    @pytest.fixture
    def digest_size(self):
        return 16

    @pytest.fixture
    def block_size(self):
        return 16


@pytest.mark.skipif(
    not mbedtls.has_feature("md4"), reason="requires MD4 support in libmbedtls"
)
class TestHashMD4(_TestHash):
    @pytest.fixture
    def algorithm(self):
        return hashlib.md4()

    @pytest.fixture
    def digest_size(self):
        return 16

    @pytest.fixture
    def block_size(self):
        return 64


@pytest.mark.skipif(
    not mbedtls.has_feature("md5"), reason="requires MD5 support in libmbedtls"
)
class TestHashMD5(_TestHash):
    @pytest.fixture
    def algorithm(self):
        return hashlib.md5()

    @pytest.fixture
    def type_(self):
        return 3

    @pytest.fixture
    def digest_size(self):
        return 16

    @pytest.fixture
    def block_size(self):
        return 64


@pytest.mark.skipif(
    not mbedtls.has_feature("sha1"),
    reason="requires SHA1 support in libmbedtls",
)
class TestHashSHA1(_TestHash):
    @pytest.fixture
    def algorithm(self):
        return hashlib.sha1()

    @pytest.fixture
    def digest_size(self):
        return 20

    @pytest.fixture
    def block_size(self):
        return 64


@pytest.mark.skipif(
    not mbedtls.has_feature("sha256"),
    reason="requires SHA256 support in libmbedtls",
)
class TestHashSHA224(_TestHash):
    @pytest.fixture
    def algorithm(self):
        return hashlib.sha224()

    @pytest.fixture
    def digest_size(self):
        return 28

    @pytest.fixture
    def block_size(self):
        return 64


@pytest.mark.skipif(
    not mbedtls.has_feature("sha256"),
    reason="requires SHA256 support in libmbedtls",
)
class TestHashSHA256(_TestHash):
    @pytest.fixture
    def algorithm(self):
        return hashlib.sha256()

    @pytest.fixture
    def digest_size(self):
        return 32

    @pytest.fixture
    def block_size(self):
        return 64


@pytest.mark.skipif(
    not mbedtls.has_feature("sha512"),
    reason="requires SHA384 support in libmbedtls",
)
class TestHashSHA384(_TestHash):
    @pytest.fixture
    def algorithm(self):
        return hashlib.sha384()

    @pytest.fixture
    def digest_size(self):
        return 48

    @pytest.fixture
    def block_size(self):
        return 128


@pytest.mark.skipif(
    not mbedtls.has_feature("sha512"),
    reason="requires SHA512 support in libmbedtls",
)
class TestHashSHA512(_TestHash):
    @pytest.fixture
    def algorithm(self):
        return hashlib.sha512()

    @pytest.fixture
    def digest_size(self):
        return 64

    @pytest.fixture
    def block_size(self):
        return 128


@pytest.mark.skipif(
    not mbedtls.has_feature("ripemd160"),
    reason="requires RIPEMD160 support in libmbedtls",
)
class TestHashRIPEMD160(_TestHash):
    @pytest.fixture
    def algorithm(self):
        return hashlib.ripemd160()

    @pytest.fixture
    def digest_size(self):
        return 20

    @pytest.fixture
    def block_size(self):
        return 64


class _TestHmac(_TestMDBase):
    @pytest.fixture(params=[0, 16])
    def key(self, request, randbytes):
        return randbytes(request.param)

    @pytest.fixture
    def buffer(self, randbytes):
        return randbytes(512)

    def test_new(self, algorithm, key, buffer):
        copy = hmaclib.new(key, buffer, algorithm.name)
        algorithm.update(buffer)
        assert algorithm.digest() == copy.digest()
        assert algorithm.hexdigest() == copy.hexdigest()


@pytest.mark.skipif(
    not mbedtls.has_feature("md2"), reason="requires MD2 support in libmbedtls"
)
class TestHmacMD2(_TestHmac):
    @pytest.fixture
    def algorithm(self, key):
        return hmaclib.md2(key)

    @pytest.fixture
    def digest_size(self):
        return 16

    @pytest.fixture
    def block_size(self):
        return 16


@pytest.mark.skipif(
    not mbedtls.has_feature("md4"), reason="requires MD4 support in libmbedtls"
)
class TestHmacMD4(_TestHmac):
    @pytest.fixture
    def algorithm(self, key):
        return hmaclib.md4(key)

    @pytest.fixture
    def digest_size(self):
        return 16

    @pytest.fixture
    def block_size(self):
        return 64


@pytest.mark.skipif(
    not mbedtls.has_feature("md5"), reason="requires MD5 support in libmbedtls"
)
class TestHmacMD5(_TestHmac):
    @pytest.fixture
    def algorithm(self, key):
        return hmaclib.md5(key)

    @pytest.fixture
    def type_(self):
        return 3

    @pytest.fixture
    def digest_size(self):
        return 16

    @pytest.fixture
    def block_size(self):
        return 64


@pytest.mark.skipif(
    not mbedtls.has_feature("sha1"),
    reason="requires SHA1 support in libmbedtls",
)
class TestHmacSHA1(_TestHmac):
    @pytest.fixture
    def algorithm(self, key):
        return hmaclib.sha1(key)

    @pytest.fixture
    def digest_size(self):
        return 20

    @pytest.fixture
    def block_size(self):
        return 64


@pytest.mark.skipif(
    not mbedtls.has_feature("sha256"),
    reason="requires SHA256 support in libmbedtls",
)
class TestHmacSHA224(_TestHmac):
    @pytest.fixture
    def algorithm(self, key):
        return hmaclib.sha224(key)

    @pytest.fixture
    def digest_size(self):
        return 28

    @pytest.fixture
    def block_size(self):
        return 64


@pytest.mark.skipif(
    not mbedtls.has_feature("sha256"),
    reason="requires SHA256 support in libmbedtls",
)
class TestHmacSHA256(_TestHmac):
    @pytest.fixture
    def algorithm(self, key):
        return hmaclib.sha256(key)

    @pytest.fixture
    def digest_size(self):
        return 32

    @pytest.fixture
    def block_size(self):
        return 64


@pytest.mark.skipif(
    not mbedtls.has_feature("sha512"),
    reason="requires SHA512 support in libmbedtls",
)
class TestHmacSHA384(_TestHmac):
    @pytest.fixture
    def algorithm(self, key):
        return hmaclib.sha384(key)

    @pytest.fixture
    def digest_size(self):
        return 48

    @pytest.fixture
    def block_size(self):
        return 128


@pytest.mark.skipif(
    not mbedtls.has_feature("sha512"),
    reason="requires SHA512 support in libmbedtls",
)
class TestHmacSHA512(_TestHmac):
    @pytest.fixture
    def algorithm(self, key):
        return hmaclib.sha512(key)

    @pytest.fixture
    def digest_size(self):
        return 64

    @pytest.fixture
    def block_size(self):
        return 128


@pytest.mark.skipif(
    not mbedtls.has_feature("ripemd160"),
    reason="requires RIPEMD160 support in libmbedtls",
)
class TestHmacRIPEMD160(_TestHmac):
    @pytest.fixture
    def algorithm(self, key):
        return hmaclib.ripemd160(key)

    @pytest.fixture
    def digest_size(self):
        return 20

    @pytest.fixture
    def block_size(self):
        return 64



import pytest

from mbedtls import secrets

# For Python 2.7
try:
    unicode
except NameError:
    unicode = str


def test_token_bytes():
    token = secrets.token_bytes()
    assert len(token) == secrets.DEFAULT_ENTROPY
    assert isinstance(token, bytes)


@pytest.mark.parametrize("nbytes", [0, 32, 256, 500])
def test_token_bytes(nbytes):
    token = secrets.token_bytes(nbytes)
    assert len(token) == nbytes
    assert isinstance(token, bytes)


def test_token_hex():
    token = secrets.token_hex()
    assert len(token) == 2 * secrets.DEFAULT_ENTROPY
    assert isinstance(token, (str, unicode))


def test_token_urlsafe():
    token = secrets.token_urlsafe()
    assert len(token) == pytest.approx(1.3 * secrets.DEFAULT_ENTROPY, rel=0.1)
    assert isinstance(token, (str, unicode))


@pytest.mark.repeat(100)
def test_choice():
    seq = tuple(range(100))
    chosen = secrets.choice(seq)
    assert chosen in seq


def test_choice_from_empty_sequence_raises_indexerror():
    with pytest.raises(IndexError):
        secrets.choice([])


def test_randbits():
    assert 0 <= secrets.randbits(32) < (1 << 32)


@pytest.mark.parametrize("upper_bound", [0, -1])
def test_randbelow_zero_raises_valueerror(upper_bound):
    with pytest.raises(ValueError):
        secrets.randbelow(upper_bound)


@pytest.mark.repeat(100)
@pytest.mark.parametrize("upper_bound", [1, 1 << 32, 1 << 128, 1 << 1024])
def test_randbelow(upper_bound):
    assert 0 <= secrets.randbelow(upper_bound) < upper_bound



# SPDX-License-Identifier: MIT
# Copyright (c) 2018, Mathias Laurin

# cython: c_api_binop_methods=True

"""Multi-precision integer library (MPI)."""


cimport mbedtls.mpi as _mpi
cimport mbedtls._random as _rnd
from libc.stdlib cimport malloc, free

import math
import numbers
from binascii import hexlify, unhexlify

import mbedtls.exceptions as _exc
import mbedtls._platform as _plt
import mbedtls._random as _rnd

try:
    long
except NameError:
    long = int


cdef _rnd.Random __rng = _rnd.default_rng()


cdef to_bytes(value):
    xx = "{0:02x}".format(value)
    return unhexlify((xx if not len(xx) % 2 else "0" + xx).encode("ascii"))


cdef from_bytes(value):
    return long(hexlify(value), 16)


cdef from_mpi_p(_mpi.mbedtls_mpi *mpi_p):
    cdef _mpi.MPI new_mpi = _mpi.MPI()
    _mpi.mbedtls_mpi_copy(&new_mpi._ctx, mpi_p)
    return new_mpi


cdef class MPI:
    """Multi-precision integer.

    This class implements `numbers.Integral`.  The representation
    of the MPI is overwritten with random bytes when the MPI is
    garbage collected.

    """
    def __init__(self, value=0):
        if isinstance(value, MPI):
            value_ = <MPI> value
            _exc.check_error(mbedtls_mpi_copy(&self._ctx, &value_._ctx))
        else:
            value = to_bytes(value)
            self._read_bytes(value)

    def __cinit__(self):
        """Initialize one MPI."""
        _mpi.mbedtls_mpi_init(&self._ctx)

    def __dealloc__(self):
        """Unallocate one MPI."""
        _exc.check_error(mbedtls_mpi_fill_random(
            &self._ctx, self._len(),
            &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
        _mpi.mbedtls_mpi_free(&self._ctx)

    def __reduce__(self):
        byteorder = "big"
        return type(self).from_bytes, (
            self.to_bytes(self._len(), byteorder),
            byteorder,
        )

    cdef size_t _len(self):
        """Return the total size in bytes."""
        return _mpi.mbedtls_mpi_size(&self._ctx)

    def _read_bytes(self, const unsigned char[:] data not None):
        if data.size == 0:
            return MPI(0)
        _exc.check_error(
            _mpi.mbedtls_mpi_read_binary(&self._ctx, &data[0], data.shape[0]))

    def __str__(self):
        return "%i" % long(self)

    def __repr__(self):
        return "%s(%i)" % (type(self).__name__, long(self))

    def bit_length(self):
        """Return the number of bits necessary to represent MPI in binary."""
        return _mpi.mbedtls_mpi_bitlen(&self._ctx)

    @classmethod
    def from_int(cls, value):
        # mbedtls_mpi_lset is 'limited' to 64 bits.
        return cls.from_bytes(to_bytes(value), byteorder="big")

    @classmethod
    def from_bytes(cls, const unsigned char[:] data, byteorder):
        assert byteorder in {"big", "little"}
        self = cls()
        self._read_bytes(data[::-1 if byteorder == "little" else 1])
        return self

    def to_bytes(self, const size_t length, byteorder):
        assert byteorder in {"big", "little"}
        cdef unsigned char* output = <unsigned char*>malloc(
            length * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(_mpi.mbedtls_mpi_write_binary(
                &self._ctx, output, length))
            return output[:length][::-1 if byteorder == "little" else 1]
        except Exception as exc:
            raise OverflowError from exc
        finally:
            free(output)

    __bytes__ = to_bytes

    @classmethod
    def prime(cls, size):
        """Return an MPI that is probably prime."""
        cdef MPI self_ = cls()
        _exc.check_error(mbedtls_mpi_gen_prime(
            &self_._ctx, size, 0,
            &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
        return self_

    def is_prime(self):
        """Miller-Rabin primality test."""
        return _exc.check_error(mbedtls_mpi_is_prime(
            &self._ctx,
            &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx)) == 0

    def __hash__(self):
        return long(self)

    def __bool__(self):
        return self != 0

    def __add__(self, other):
        if not all((isinstance(self, numbers.Integral),
                    isinstance(other, numbers.Integral))):
            return NotImplemented
        cdef MPI self_ = MPI(self)
        cdef MPI other_ = MPI(other)
        cdef MPI result = MPI()
        _exc.check_error(mbedtls_mpi_add_mpi(
            &result._ctx, &self_._ctx, &other_._ctx))
        return result

    def __neg__(self):
        raise TypeError("negative value")

    def __pos__(self):
        return self

    def __sub__(self, other):
        if not all((isinstance(self, numbers.Integral),
                    isinstance(other, numbers.Integral))):
            return NotImplemented
        cdef MPI self_ = MPI(self)
        cdef MPI other_ = MPI(other)
        cdef MPI result = MPI()
        _exc.check_error(mbedtls_mpi_sub_mpi(
            &result._ctx, &self_._ctx, &other_._ctx))
        return result

    def __mul__(self, other):
        if not all((isinstance(self, numbers.Integral),
                    isinstance(other, numbers.Integral))):
            return NotImplemented
        cdef MPI self_ = MPI(self)
        cdef MPI other_ = MPI(other)
        cdef MPI result = MPI()
        _exc.check_error(mbedtls_mpi_mul_mpi(
            &result._ctx, &self_._ctx, &other_._ctx))
        return result

    def __truediv__(self, other):
        return NotImplemented

    def __pow__(MPI self, exponent, modulus):
        if not isinstance(exponent, numbers.Integral):
            return TypeError("exponent should be an integer")
        if not isinstance(modulus, numbers.Integral):
            return TypeError("modulus should be an integer")
        if exponent < 0:
            raise ValueError("exponent must be greater than zero")
        cdef MPI result = MPI()
        cdef MPI exponent_ = MPI(exponent)
        cdef MPI modulus_ = MPI(modulus)
        _exc.check_error(
            mbedtls_mpi_exp_mod(
                &result._ctx, &self._ctx, &exponent_._ctx, &modulus_._ctx, NULL
            )
        )
        return result

    def __abs__(self):
        # Negative values are not supported.
        return self

    def __eq__(MPI self, other):
        if not isinstance(other, numbers.Integral):
            return NotImplemented
        cdef MPI other_ = MPI(other)
        return mbedtls_mpi_cmp_mpi(&self._ctx, &other_._ctx) == 0

    def __float__(self):
        return float(long(self))

    def __trunc__(self):
        return self

    def __floor__(self):
        return self

    def __ceil__(self):
        return self

    def __round__(self, ndigits=None):
        return self

    def __divmod__(self, other):
        if not all((isinstance(self, numbers.Integral),
                    isinstance(other, numbers.Integral))):
            return NotImplemented
        cdef MPI self_ = MPI(self)
        cdef MPI other_ = MPI(other)
        cdef MPI quotient = MPI()
        cdef MPI rest = MPI()
        _exc.check_error(mbedtls_mpi_div_mpi(
            &quotient._ctx, &rest._ctx, &self_._ctx, &other_._ctx))
        return quotient, rest

    def __floordiv__(self, other):
        return divmod(self, other)[0]

    def __mod__(self, other):
        if not all((isinstance(self, numbers.Integral),
                    isinstance(other, numbers.Integral))):
            return NotImplemented
        cdef MPI self_ = MPI(self)
        cdef MPI other_ = MPI(other)
        cdef MPI result = MPI()
        _exc.check_error(mbedtls_mpi_mod_mpi(
            &result._ctx, &self_._ctx, &other_._ctx))
        return result

    def __lt__(MPI self, other):
        if not isinstance(other, numbers.Integral):
            return NotImplemented
        cdef MPI other_ = MPI(other)
        return mbedtls_mpi_cmp_mpi(&self._ctx, &other_._ctx) == -1

    def __le__(MPI self, other):
        return self < other or self == other

    def __gt__(MPI self, other):
        return not self <= other

    def __ge__(MPI self, other):
        return self > other or self == other

    def __complex__(self):
        return complex(float(self))

    def real(self):
        return self

    def imag(self):
        return 0

    def conjugate(self):
        return self

    def __int__(self):
        n = self._len()
        if n:
            return from_bytes(self.to_bytes(n, byteorder="big"))
        else:
            return 0

    def __index__(self):
        return long(self)

    def __lshift__(MPI self, other):
        _exc.check_error(mbedtls_mpi_shift_l(&self._ctx, long(other)))
        return self

    def __rshift__(MPI self, other):
        _exc.check_error(mbedtls_mpi_shift_r(&self._ctx, long(other)))
        return self

    def __and__(MPI self, other):
        if not isinstance(other, MPI):
            other = MPI(other)
        cdef size_t size = long(
            math.ceil(max(self.bit_length(), other.bit_length()) / 8)
        )
        self_bin = bytearray(self.to_bytes(size, "big"))
        other_bin = bytearray(other.to_bytes(size, "big"))
        output = bytearray(size)
        cdef size_t ii
        for ii in range(size):
            output[ii] = self_bin[ii] & other_bin[ii]
        result = MPI.from_bytes(output, "big")
        _plt.zeroize(self_bin)
        _plt.zeroize(other_bin)
        _plt.zeroize(output)
        return result

    def __xor__(MPI self, other):
        if not isinstance(other, MPI):
            other = MPI(other)
        cdef size_t size = long(
            math.ceil(max(self.bit_length(), other.bit_length()) / 8)
        )
        self_bin = bytearray(self.to_bytes(size, "big"))
        other_bin = bytearray(other.to_bytes(size, "big"))
        output = bytearray(size)
        cdef size_t ii
        for ii in range(size):
            output[ii] = self_bin[ii] ^ other_bin[ii]
        result = MPI.from_bytes(output, "big")
        _plt.zeroize(self_bin)
        _plt.zeroize(other_bin)
        _plt.zeroize(output)
        return result

    def __or__(MPI self, other):
        if not isinstance(other, MPI):
            other = MPI(other)
        cdef size_t size = long(
            math.ceil(max(self.bit_length(), other.bit_length()) / 8)
        )
        self_bin = bytearray(self.to_bytes(size, "big"))
        other_bin = bytearray(other.to_bytes(size, "big"))
        output = bytearray(size)
        cdef size_t ii
        for ii in range(size):
            output[ii] = self_bin[ii] | other_bin[ii]
        result = MPI.from_bytes(output, "big")
        _plt.zeroize(self_bin)
        _plt.zeroize(other_bin)
        _plt.zeroize(output)
        return result

    def __invert__(self):
        raise TypeError("negative value")

    @property
    def numerator(self):
        return self

    @property
    def denominator(self):
        return 1


numbers.Integral.register(MPI)



# SPDX-License-Identifier: MIT
# Copyright (c) 2016, Elaborated Networks GmbH
# Copyright (c) 2018, Mathias Laurin

"""Public key (PK) library.

The library handles RSA certificates and ECC (elliptic curve
cryptography).

The RSA and ECC classes offer compatible APIs.  They may be used
interchangeably.

ECDHServer and ECDHClient should be used for ephemeral Elliptic Curve
Diffie-Hellman exchange.

"""


from libc.stdlib cimport malloc, free
from libc.string cimport memset

cimport mbedtls.mpi as _mpi
cimport mbedtls.pk as _pk
cimport mbedtls._random as _rnd

import enum
from collections import namedtuple
from functools import partial

import mbedtls._random as _rnd
import mbedtls.exceptions as _exc
from mbedtls.hashlib import new as _new_hash


__all__ = ("check_pair", "get_supported_ciphers", "get_supported_curves",
           "Curve", "RSA", "ECC", "DHServer", "DHClient",
           "ECDHServer", "ECDHClient", "ECDHNaive")


CIPHER_NAME = (
    b"NONE",
    b"RSA",
    b"EC",
    b"EC_DH",
    b"ECDSA",
    # b"RSA_ALT",
    # b"RSASSA_PSS",
)


class CipherType(enum.Enum):
    NONE = _pk.MBEDTLS_PK_NONE
    RSA = _pk.MBEDTLS_PK_RSA
    ECKEY = _pk.MBEDTLS_PK_ECKEY
    ECKEY_DH = _pk.MBEDTLS_PK_ECKEY_DH
    ECDSA = _pk.MBEDTLS_PK_ECDSA
    RSA_ALT = _pk.MBEDTLS_PK_RSA_ALT
    RSASSA_PSS = _pk.MBEDTLS_PK_RSASSA_PSS


KeyPair = namedtuple("KeyPair", ["private", "public"])


class Curve(bytes, enum.Enum):

    """Elliptic curves."""

    SECP192R1 = b'secp192r1'
    SECP224R1 = b'secp224r1'
    SECP256R1 = b'secp256r1'
    SECP384R1 = b'secp384r1'
    SECP521R1 = b'secp521r1'
    BRAINPOOLP256R1 = b'brainpoolP256r1'
    BRAINPOOLP384R1 = b'brainpoolP384r1'
    BRAINPOOLP512R1 = b'brainpoolP512r1'
    SECP192K1 = b'secp192k1'
    SECP224K1 = b'secp224k1'
    SECP256K1 = b'secp256k1'
    CURVE25519 = b'x25519'
    CURVE448 = b'x448'


# The following calculations come from mbedtls/library/pkwrite.c.
RSA_PUB_DER_MAX_BYTES = 38 + 2 * _mpi.MBEDTLS_MPI_MAX_SIZE
MPI_MAX_SIZE_2 = _mpi.MBEDTLS_MPI_MAX_SIZE // 2 + _mpi.MBEDTLS_MPI_MAX_SIZE % 2
RSA_PRV_DER_MAX_BYTES = 47 + 3 * _mpi.MBEDTLS_MPI_MAX_SIZE + 5 * MPI_MAX_SIZE_2

ECP_PUB_DER_MAX_BYTES = 30 + 2 * _pk.MBEDTLS_ECP_MAX_BYTES
ECP_PRV_DER_MAX_BYTES = 29 + 3 * _pk.MBEDTLS_ECP_MAX_BYTES

PUB_DER_MAX_BYTES = max(RSA_PUB_DER_MAX_BYTES, ECP_PUB_DER_MAX_BYTES)
PRV_DER_MAX_BYTES = max(RSA_PRV_DER_MAX_BYTES, ECP_PRV_DER_MAX_BYTES)

del RSA_PUB_DER_MAX_BYTES, MPI_MAX_SIZE_2, RSA_PRV_DER_MAX_BYTES
del ECP_PUB_DER_MAX_BYTES, ECP_PRV_DER_MAX_BYTES


cpdef check_pair(CipherBase pub, CipherBase pri):
    """Check if a public-private pair of keys matches."""
    return _pk.mbedtls_pk_check_pair(&pub._ctx, &pri._ctx) == 0


def _type_from_name(name):
    return {name: n for n, name in enumerate(CIPHER_NAME)}.get(name, 0)


cpdef get_supported_ciphers():
    return CIPHER_NAME


def get_supported_curves():
    """Return the list of supported curves in order of preference."""
    cdef const mbedtls_ecp_curve_info* info = _pk.mbedtls_ecp_curve_list()
    names, idx = [], 0
    while info[idx].name != NULL:
        names.append(Curve(bytes(info[idx].name)))
        idx += 1
    return names


cdef curve_name_to_grp_id(curve):
    cdef const mbedtls_ecp_curve_info* info = _pk.mbedtls_ecp_curve_list()
    idx = 0
    while info[idx].name != NULL:
        if info[idx].name == curve:
            return info.grp_id
        idx += 1
    raise LookupError(curve.decode("ascii") + " not found")


cdef _rnd.Random __rng = _rnd.default_rng()


def _get_md_alg(digestmod):
    """Return the hash object.

    Arguments:
        digestmod: The digest name or digest constructor for the
            Cipher object to use.  It supports any name suitable to
            `mbedtls.hash.new()`.

    """
    # `digestmod` handling below is adapted from CPython's
    # `hmac.py`.
    if callable(digestmod):
        return digestmod
    elif isinstance(digestmod, (str, unicode)):
        return partial(_new_hash, digestmod)
    else:
        raise TypeError("a valid digestmod is required, got %r" % digestmod)


cdef class CipherBase:
    """Base class to RSA and ECC ciphers.

    Parameters:
        name (bytes): The cipher name known to mbed TLS.
        key (bytes, optional): A key (public or private half).
        password (bytes, optional): The password for the key.

    """
    def __init__(self,
                 name,
                 const unsigned char[:] key=None,
                 const unsigned char[:] password=None):
        _exc.check_error(_pk.mbedtls_pk_setup(
            &self._ctx,
            _pk.mbedtls_pk_info_from_type(
                _type_from_name(name)
            )
        ))
        if key is None or key.size == 0:
            return
        mbedtls_pk_free(&self._ctx)  # The context must be reset on entry.
        try:
            if password is None or password.size == 0:
                _exc.check_error(_pk.mbedtls_pk_parse_key(
                    &self._ctx, &key[0], key.size, NULL, 0
                ))
            else:
                _exc.check_error(_pk.mbedtls_pk_parse_key(
                    &self._ctx, &key[0], key.size, &password[0], password.size
                ))
        except _exc.TLSError:
            _exc.check_error(_pk.mbedtls_pk_parse_public_key(
                &self._ctx, &key[0], key.size))

    def __cinit__(self):
        """Initialize the context."""
        _pk.mbedtls_pk_init(&self._ctx)

    def __reduce__(self):
        key = self.export_key()
        if not key:
            key = self.export_public_key()
        return type(self).from_buffer, (key,)

    def __dealloc__(self):
        """Free and clear the context."""
        _pk.mbedtls_pk_free(&self._ctx)

    def __hash__(self):
        if self._has_private():
            return hash(self.export_key(format="DER"))
        else:
            return hash(self.export_public_key(format="DER"))

    def __eq__(self, other):
        if type(other) is not type(self):
            try:
                other = other.encode("ascii")
            except:
                pass
            try:
                other = type(self).from_buffer(other)
            except:
                return False
        if self._has_private() or other._has_private():
            return other.export_key() == self.export_key()
        elif not (self._has_private() and other._has_private()):
            return other.export_public_key() == self.export_public_key()

    @classmethod
    def from_buffer(cls, const unsigned char[:] key not None):
        """Import a key (public or private half).

        The public half is generated upon importing a private key.

        Arguments:
            key (bytes): The key in PEM or DER format.
            password (bytes, optional): The password for
                password-protected private keys.

        """
        raise NotImplementedError

    @classmethod
    def from_DER(cls, const unsigned char[:] key not None):
        return cls.from_buffer(key)

    @classmethod
    def from_PEM(cls, key):
        """Import a key (public or private half)."""
        return cls.from_buffer(key.encode("ascii"))

    @property
    def _type(self):
        """Return the type of the cipher."""
        return _pk.mbedtls_pk_get_type(&self._ctx)
    
    @property
    def name(self):
        """Return the name of the cipher."""
        return _pk.mbedtls_pk_get_name(&self._ctx)

    @property
    def _bitlen(self):
        """Return the size of the key, in bits."""
        return _pk.mbedtls_pk_get_bitlen(&self._ctx)

    @property
    def key_size(self):
        """Return the size of the key, in bytes."""
        return _pk.mbedtls_pk_get_len(&self._ctx)

    def _has_private(self):
        """Return `True` if the key contains a valid private half."""
        raise NotImplementedError

    def _has_public(self):
        """Return `True` if the key contains a valid public half."""
        raise NotImplementedError

    def sign(self,
             const unsigned char[:] message not None,
             digestmod=None):
        """Make signature, including padding if relevant.

        Arguments:
            message (bytes): The message to sign.
            digestmod (optional): The digest name or digest constructor.

        Return:
            bytes or None: The signature or None if the cipher does not
                contain a private key.

        """
        if digestmod is None:
            digestmod = 'sha256'
        if not self._has_private():
            return None
        md_alg = _get_md_alg(digestmod)(message)
        cdef const unsigned char[:] hash_ = md_alg.digest()
        cdef size_t sig_len = 0
        cdef unsigned char* output = <unsigned char*>malloc(
            _mpi.MBEDTLS_MPI_MAX_SIZE * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(_pk.mbedtls_pk_sign(
                &self._ctx, md_alg._type,
                &hash_[0], hash_.size,
                &output[0], &sig_len,
                &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
            assert sig_len != 0
            return output[:sig_len]
        finally:
            free(output)

    def verify(self,
               const unsigned char[:] message not None,
               const unsigned char[:] signature not None,
               digestmod=None):
        """Verify signature, including padding if relevant.

        Arguments:
            message (bytes): The message to sign.
            signature (bytes): The signature to verify.
            digestmod (optional): The digest name or digest constructor.

        Return:
            bool: True if the verification passed, False otherwise.

        """
        if signature.size == 0:
            return False
        if digestmod is None:
            digestmod = 'sha256'
        md_alg = _get_md_alg(digestmod)(message)
        cdef const unsigned char[:] hash_ = md_alg.digest()
        return _pk.mbedtls_pk_verify(
            &self._ctx, md_alg._type,
            &hash_[0], hash_.size,
            &signature[0], signature.size) == 0

    def encrypt(self, const unsigned char[:] message not None):
        """Encrypt message (including padding if relevant).

        Arguments:
            message (bytes): Message to encrypt.

        """
        if message.size == 0:
            message = b"\0"
        cdef size_t olen = 0
        cdef unsigned char* output = <unsigned char*>malloc(
            _mpi.MBEDTLS_MPI_MAX_SIZE // 2 * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(_pk.mbedtls_pk_encrypt(
                &self._ctx, &message[0], message.size,
                output, &olen, self.key_size,
                &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
            return output[:olen]
        finally:
            free(output)

    def decrypt(self, const unsigned char[:] message not None):
        """Decrypt message (including padding if relevant).

        Arguments:
            message (bytes): Message to decrypt.

        """
        if message.size == 0:
            message = b"\0"
        cdef size_t olen = 0
        cdef unsigned char* output = <unsigned char*>malloc(
            _mpi.MBEDTLS_MPI_MAX_SIZE // 2 * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(_pk.mbedtls_pk_decrypt(
                &self._ctx, &message[0], message.size,
                output, &olen, self.key_size,
                &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
            return output[:olen]
        finally:
            free(output)

    def generate(self):
        """Generate a keypair.

        Return:
            (bytes): The private key in DER format.

        """
        raise NotImplementedError

    def _private_to_DER(self):
        if not self._has_private():
            return b""
        cdef int olen
        cdef size_t osize = PRV_DER_MAX_BYTES
        cdef unsigned char *output = <unsigned char *>malloc(
            osize * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            olen = _exc.check_error(
                _pk.mbedtls_pk_write_key_der(&self._ctx, output, osize))
            return output[osize - olen:osize]
        finally:
            free(output)

    def _private_to_PEM(self):
        if not self._has_private():
            return ""
        cdef size_t osize = PRV_DER_MAX_BYTES * 4 // 3 + 100
        cdef unsigned char *output = <unsigned char *>malloc(
            osize * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        memset(output, 0, osize)
        try:
            _exc.check_error(
                _pk.mbedtls_pk_write_key_pem(&self._ctx, output, osize))
            return output[0:osize].decode("ascii")
        finally:
            free(output)

    def export_key(self, format="DER"):
        """Return the private key.

        If no key is present, return a falsy value.

        Args:
            format (str): One of "DER" or "PEM".

        """
        if format == "DER":
            return self._private_to_DER()
        if format == "PEM":
            return self._private_to_PEM()
        raise ValueError(format)

    def _public_to_DER(self):
        if not self._has_public():
            return b""
        cdef int olen
        cdef size_t osize = PRV_DER_MAX_BYTES
        cdef unsigned char *output = <unsigned char *>malloc(
            osize * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            olen = _exc.check_error(
                _pk.mbedtls_pk_write_pubkey_der(&self._ctx, output, osize))
            return output[osize - olen:osize]
        finally:
            free(output)

    def _public_to_PEM(self):
        if not self._has_public():
            return ""
        cdef size_t osize = PRV_DER_MAX_BYTES * 4 // 3 + 100
        cdef unsigned char *output = <unsigned char *>malloc(
            osize * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        memset(output, 0, osize)
        try:
            _exc.check_error(
                _pk.mbedtls_pk_write_pubkey_pem(&self._ctx, output, osize))
            return output[0:osize].decode("ascii")
        finally:
            free(output)

    def export_public_key(self, format="DER"):
        """Return the public key.

        If no key is present, return a falsy value.

        Args:
            format (str): One of "DER" or "PEM".

        """
        if format == "DER":
            return self._public_to_DER()
        if format == "PEM":
            return self._public_to_PEM()
        raise ValueError(format)

    def to_PEM(self):
        """Return the RSA in PEM format.

        Warning:
            This function is obsolete.

        Return:
            tuple(str, str): The private key and the public key.

        See Also:
            export_key(), export_public_key()

        """
        return KeyPair(self.export_key("PEM"), self.export_public_key("PEM"))

    def __str__(self):
        return self.export_key(format="PEM")

    def to_DER(self):
        """Return the RSA in DER format.

        Warning:
            This function is obsolete.

        Return:
            tuple(bytes, bytes): The private key and the public key.

        See Also:
            export_key(), export_public_key()

        """
        return KeyPair(self.export_key("DER"), self.export_public_key("DER"))

    def to_bytes(self):
        """Return the private key in DER format."""
        return self.export_key(format="DER")

    def __bytes__(self):
        return self.to_bytes()


cdef class RSA(CipherBase):

    """RSA public-key cryptosystem."""

    def __init__(self,
                 const unsigned char[:] key=None,
                 const unsigned char[:] password=None):
        super().__init__(b"RSA", key, password)

    @classmethod
    def from_buffer(cls, const unsigned char[:] key):
        """Import a key (public or private half).

        The public half is generated upon importing a private key.

        Arguments:
            key (bytes): The key in PEM or DER format.
            password (bytes, optional): The password for
                password-protected private keys.

        """
        return cls(key)

    def _has_private(self):
        """Return `True` if the key contains a valid private half."""
        return _pk.mbedtls_rsa_check_privkey(_pk.mbedtls_pk_rsa(self._ctx)) == 0

    def _has_public(self):
        """Return `True` if the key contains a valid public half."""
        return _pk.mbedtls_rsa_check_pubkey(_pk.mbedtls_pk_rsa(self._ctx)) == 0

    def generate(self, unsigned int key_size=2048, int exponent=65537):
        """Generate an RSA keypair.

        Arguments:
            key_size (unsigned int): size in bits.
            exponent (int): public RSA exponent.

        Return:
            (bytes): The private key in DER format.

        """
        _exc.check_error(_pk.mbedtls_rsa_gen_key(
            _pk.mbedtls_pk_rsa(self._ctx), &_rnd.mbedtls_ctr_drbg_random,
            &__rng._ctx, key_size, exponent))
        return self.export_key("DER")


cdef class ECPoint:

    """A point on the elliptic curve."""
    def __init__(self, x, y, z):
        cdef _mpi.MPI _x = _mpi.MPI(x)
        cdef _mpi.MPI _y = _mpi.MPI(y)
        cdef _mpi.MPI _z = _mpi.MPI(z)
        _mpi.mbedtls_mpi_copy(&self._ctx.X, &_x._ctx)
        _mpi.mbedtls_mpi_copy(&self._ctx.Y, &_y._ctx)
        _mpi.mbedtls_mpi_copy(&self._ctx.Z, &_z._ctx)

    def __cinit__(self):
        """Initialize the context."""
        _pk.mbedtls_ecp_point_init(&self._ctx)

    def __dealloc__(self):
        """Free and clear the context."""
        _pk.mbedtls_ecp_point_free(&self._ctx)

    def __reduce__(self):
        return type(self), (self.x, self.y, self.z)

    @property
    def x(self):
        """Return the X coordinate."""
        try:
            return _mpi.from_mpi_p(&self._ctx.X)
        except ValueError:
            return _mpi.MPI()

    @property
    def y(self):
        """Return the Y coordinate."""
        try:
            return _mpi.from_mpi_p(&self._ctx.Y)
        except ValueError:
            return _mpi.MPI()

    @property
    def z(self):
        """Return the Z coordinate."""
        try:
            return _mpi.from_mpi_p(&self._ctx.Z)
        except ValueError:
            return _mpi.MPI()

    def __str__(self):
        return "%s(%i, %i, %i)" % (
            type(self).__name__, self.x, self.y, self.z
        )

    def __repr__(self):
        return "%s(%r, %r, %r)" % (
            type(self).__name__, self.x, self.y, self.z
        )

    def __eq__(self, other):
        if other == 0:
            return _pk.mbedtls_ecp_is_zero(&self._ctx) == 1
        elif type(other) is type(self):
            c_other = <ECPoint> other
            return _pk.mbedtls_ecp_point_cmp(&self._ctx, &c_other._ctx) == 0
        return NotImplemented

    def __hash__(self):
        return hash((self.x, self.y, self.z))

    def __bool__(self):
        return _pk.mbedtls_ecp_is_zero(&self._ctx) == 0


cdef class ECC(CipherBase):

    """Elliptic-curve cryptosystems.

    Args:
        (Curve, optional): A curve returned by `get_supported_curves()`.

    See Also:
        get_supported_curves()

    """
    def __init__(self,
                 curve=None,
                 const unsigned char[:] key=None,
                 const unsigned char[:] password=None):
        super().__init__(b"EC", key, password)
        if curve is None:
            curve = get_supported_curves()[0]
        self.curve = curve

    @classmethod
    def from_buffer(cls, const unsigned char[:] key not None):
        """Import a key (public or private half).

        The public half is generated upon importing a private key.

        Arguments:
            key (bytes): The key in PEM or DER format.
            password (bytes, optional): The password for
                password-protected private keys.

        """
        return cls(None, key)

    def _has_private(self):
        """Return `True` if the key contains a valid private half."""
        cdef const mbedtls_ecp_keypair* ecp = _pk.mbedtls_pk_ec(self._ctx)
        return _mpi.mbedtls_mpi_cmp_mpi(&ecp.d, &_mpi.MPI()._ctx) != 0

    def _has_public(self):
        """Return `True` if the key contains a valid public half."""
        cdef mbedtls_ecp_keypair* ecp = _pk.mbedtls_pk_ec(self._ctx)
        return not _pk.mbedtls_ecp_is_zero(&ecp.Q)

    def generate(self):
        """Generate an EC keypair.

        Return:
            (bytes): The private key in DER format.

        """
        grp_id = curve_name_to_grp_id(self.curve)
        if grp_id is None:
            raise ValueError(self.curve)
        _exc.check_error(_pk.mbedtls_ecp_gen_key(
            grp_id, _pk.mbedtls_pk_ec(self._ctx),
            &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
        return self.export_key("DER")

    def _private_to_num(self):
        try:
            return _mpi.from_mpi_p(&_pk.mbedtls_pk_ec(self._ctx).d)
        except ValueError:
            return _mpi.MPI()

    def export_key(self, format="DER"):
        """Return the private key.

        If not key is present, return a falsy value.

        Args:
            format (str): One of "DER", "PEM", or "NUM".

        """
        if format == "NUM":
            return self._private_to_num()
        return super().export_key(format)

    def _public_to_point(self):
        point = ECPoint(0, 0, 0)
        _pk.mbedtls_ecp_copy(&point._ctx, &_pk.mbedtls_pk_ec(self._ctx).Q)
        return point

    def export_public_key(self, format="DER"):
        """Return the public key.

        If no key is present, return a falsy value.

        Args:
            format (str): One of "DER", "PEM", or "POINT".

        """
        if format == "POINT":
            return self._public_to_point()
        return super().export_public_key(format)

    def to_ECDH_server(self):
        """Return an ECDH server initialized with this context."""
        ecdh = ECDHServer(self.curve)
        _exc.check_error(_pk.mbedtls_ecdh_get_params(
            &ecdh._ctx, _pk.mbedtls_pk_ec(self._ctx), MBEDTLS_ECDH_OURS))
        return ecdh

    def to_ECDH_client(self):
        """Return an ECDH client initialized with this context."""
        ecdh = ECDHClient(self.curve)
        _exc.check_error(_pk.mbedtls_ecdh_get_params(
            &ecdh._ctx, _pk.mbedtls_pk_ec(self._ctx), MBEDTLS_ECDH_THEIRS))
        return ecdh


cdef class DHBase:

    """Base class to DH key exchange: client and server.

    Args:
        modulus (int): The prime modulus P.
        generator (int): The generator G, a primitive root modulo P.

    See Also:
        DHServer, DHClient: The derived classes.

    """
    def __init__(self, modulus, generator):
        super().__init__()
        _exc.check_error(_mpi.mbedtls_mpi_copy(
            &self._ctx.P, &_mpi.MPI(modulus)._ctx))
        _exc.check_error(_mpi.mbedtls_mpi_copy(
            &self._ctx.G, &_mpi.MPI(generator)._ctx))

    def __cinit__(self):
        """Initialize the context."""
        _pk.mbedtls_dhm_init(&self._ctx)

    def __dealloc__(self):
        """Free and clear the context."""
        _pk.mbedtls_dhm_free(&self._ctx)

    def __getstate__(self):
        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")

    @property
    def key_size(self):
        """Return the size of the key, in bytes."""
        return _mpi.mbedtls_mpi_size(&self._ctx.P)

    @property
    def modulus(self):
        """Return the prime modulus, P."""
        return _mpi.from_mpi_p(&self._ctx.P)

    @property
    def generator(self):
        """Return the generator, G."""
        return _mpi.from_mpi_p(&self._ctx.G)

    @property
    def _secret(self):
        """Return the secret (int)."""
        return _mpi.from_mpi_p(&self._ctx.X)

    @property
    def shared_secret(self):
        """The shared secret (int).

        The shared secret is 0 if the TLS handshake is not finished.

        """
        try:
            return _mpi.from_mpi_p(&self._ctx.K)
        except ValueError:
            return _mpi.MPI()

    def generate_secret(self):
        """Generate the shared secret."""
        cdef _mpi.MPI mpi
        cdef unsigned char* output = <unsigned char*>malloc(
            _mpi.MBEDTLS_MPI_MAX_SIZE * sizeof(unsigned char))
        cdef size_t olen = 0
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(mbedtls_dhm_calc_secret(
                &self._ctx, &output[0], _mpi.MBEDTLS_MPI_MAX_SIZE, &olen,
                &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
            mpi = _mpi.MPI()
            _mpi.mbedtls_mpi_read_binary(&mpi._ctx, &output[0], olen)
            return mpi
        finally:
            free(output)

    
cdef class DHServer(DHBase):

    """The server side of the DH key exchange."""

    def generate(self):
        """Generate a public key.

        Return:
            bytes: A TLS ServerKeyExchange payload.

        """
        cdef unsigned char* output = <unsigned char*>malloc(
            _mpi.MBEDTLS_MPI_MAX_SIZE * sizeof(unsigned char))
        cdef size_t olen = 0
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(_pk.mbedtls_dhm_make_params(
                &self._ctx, self.key_size, &output[0], &olen,
                &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
            assert olen != 0
            return output[:olen]
        finally:
            free(output)

    def import_CKE(self, const unsigned char[:] buffer not None):
        """Read the ClientKeyExchange payload."""
        if buffer.size == 0:
            buffer = b"\0"
        _exc.check_error(_pk.mbedtls_dhm_read_public(
            &self._ctx, &buffer[0], buffer.size))


cdef class DHClient(DHBase):

    """The client side of the DH key exchange."""

    def generate(self):
        """Generate the public key.

        Return:
            bytes: The byte representation (big endian) of: G^X mod P.

        """
        cdef _mpi.MPI mpi
        cdef unsigned char* output = <unsigned char*>malloc(
            _mpi.MBEDTLS_MPI_MAX_SIZE * sizeof(unsigned char))
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(_pk.mbedtls_dhm_make_public(
                &self._ctx, self.key_size, &output[0], self.key_size,
                &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
            mpi = _mpi.from_mpi_p(&self._ctx.GX)
            return mpi.to_bytes(
                _mpi.mbedtls_mpi_size(&mpi._ctx), "big")
        finally:
            free(output)

    def import_SKE(self, const unsigned char[:] buffer not None):
        """Read the ServerKeyExchange payload."""
        if buffer.size == 0:
            raise ValueError("SKE is empty")
        # Const away cast is safe because the pointer is used as
        # a read-only index in `dhm_read_bignum()`.
        cdef unsigned char* first = <unsigned char*> &buffer[0]
        cdef const unsigned char* end = &buffer[-1] + 1
        _exc.check_error(_pk.mbedtls_dhm_read_params(&self._ctx, &first, end))


cdef class ECDHBase:

    """Base class to ECDH(E) key exchange: client and server.

    Args:
        (Curve, optional): A curve returned by `get_supported_curves()`.

    See Also:
        ECDHServer, ECDHClient: The derived class.
        get_supported_curves()

    """
    def __init__(self, curve=None):
        super().__init__()
        if curve is None:
            curve = get_supported_curves()[0]
        self.curve = Curve(curve)
        _exc.check_error(mbedtls_ecp_group_load(
            &self._ctx.grp, curve_name_to_grp_id(self.curve)))

    def __cinit__(self):
        """Initialize the context."""
        _pk.mbedtls_ecdh_init(&self._ctx)

    def __dealloc__(self):
        """Free and clear the context."""
        _pk.mbedtls_ecdh_free(&self._ctx)

    def __getstate__(self):
        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")

    def _has_private(self):
        """Return `True` if the key contains a valid private half."""
        return _mpi.mbedtls_mpi_cmp_mpi(&self._ctx.d, &_mpi.MPI()._ctx) != 0

    def _has_public(self):
        """Return `True` if the key contains a valid public half."""
        return not _pk.mbedtls_ecp_is_zero(&self._ctx.Q)

    def _has_peers_public(self):
        """Return `True` if the peer's key is present."""
        return not _pk.mbedtls_ecp_is_zero(&self._ctx.Qp)

    @property
    def public_key(self):
        """The public key (ECPoint)"""
        ecp = ECPoint(0, 0, 0)
        _exc.check_error(_pk.mbedtls_ecp_copy(&ecp._ctx, &self._ctx.Q))
        return ecp

    @property
    def private_key(self):
        """The private key (int)"""
        return _mpi.from_mpi_p(&self._ctx.d)

    @property
    def peers_public_key(self):
        """Peer's public key (ECPoint)"""
        ecp = ECPoint(0, 0, 0)
        _exc.check_error(_pk.mbedtls_ecp_copy(&ecp._ctx, &self._ctx.Qp))
        return ecp

    @property
    def shared_secret(self):
        """The shared secret (int).

        The shared secret is 0 if the TLS handshake is not finished.

        """
        try:
            return _mpi.from_mpi_p(&self._ctx.z)
        except ValueError:
            return _mpi.MPI()

    def generate_secret(self):
        """Generate the shared secret."""
        cdef _mpi.MPI mpi = _mpi.MPI()
        cdef unsigned char* output = <unsigned char*>malloc(
            _mpi.MBEDTLS_MPI_MAX_SIZE * sizeof(unsigned char))
        cdef size_t olen = 0
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(mbedtls_ecdh_calc_secret(
                &self._ctx, &olen, &output[0], _mpi.MBEDTLS_MPI_MAX_SIZE,
                &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
            assert olen != 0
            _mpi.mbedtls_mpi_read_binary(&mpi._ctx, &output[0], olen)
            return mpi
        finally:
            free(output)

    def generate_public_key(self):
        """Generate public key from a private key."""
        _exc.check_error(_pk.mbedtls_ecp_mul(
            &self._ctx.grp, &self._ctx.Q, &self._ctx.d, &self._ctx.grp.G,
            &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))

    @property
    def _private_key(self):
        return self.private_key

    @_private_key.setter
    def _private_key(self, priv):
        cdef _mpi.MPI c_priv = _mpi.MPI(priv)
        _mpi.mbedtls_mpi_copy(&self._ctx.d, &c_priv._ctx)

    @property
    def _public_key(self):
        return self.public_key

    @_public_key.setter
    def _public_key(self, ECPoint ecp):
        _exc.check_error(_pk.mbedtls_ecp_copy(&self._ctx.Q, &ecp._ctx))

    @property
    def _peer_public_key(self):
        return self.peers_public_key

    @_peer_public_key.setter
    def _peer_public_key(self, ECPoint ecp):
        _exc.check_error(_pk.mbedtls_ecp_copy(&self._ctx.Qp, &ecp._ctx))


cdef class ECDHServer(ECDHBase):

    """The server side of the ECDH key exchange.

    Args:
        (Curve, optional): A curve returned by `get_supported_curves()`.

    """
    def generate(self):
        """Generate a public key.

        Return:
            bytes: A TLS ServerKeyExchange payload.

        """
        cdef unsigned char* output = <unsigned char*>malloc(
            _mpi.MBEDTLS_MPI_MAX_SIZE * sizeof(unsigned char))
        cdef size_t olen = 0
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(_pk.mbedtls_ecdh_make_params(
                &self._ctx, &olen, &output[0], _mpi.MBEDTLS_MPI_MAX_SIZE,
                &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
            assert olen != 0
            return output[:olen]
        finally:
            free(output)

    def import_CKE(self, const unsigned char[:] buffer not None):
        """Read the ClientKeyExchange payload."""
        if buffer.size == 0:
            buffer = b"\0"
        _exc.check_error(_pk.mbedtls_ecdh_read_public(
            &self._ctx, &buffer[0], buffer.size))


cdef class ECDHClient(ECDHBase):

    """The client side of the ephemeral ECDH key exchange.

    Args:
        (Curve, optional): A curve returned by `get_supported_curves()`.

    """
    def generate(self):
        """Generate a public key.

        Return:
            bytes: A TLS ClientKeyExchange payload.

        """
        cdef unsigned char* output = <unsigned char*>malloc(
            _mpi.MBEDTLS_MPI_MAX_SIZE * sizeof(unsigned char))
        cdef size_t olen = 0
        if not output:
            raise MemoryError()
        try:
            _exc.check_error(_pk.mbedtls_ecdh_make_public(
                &self._ctx, &olen, &output[0], _mpi.MBEDTLS_MPI_MAX_SIZE,
                &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
            assert olen != 0
            return output[:olen]
        finally:
            free(output)

    def import_SKE(self, const unsigned char[:] buffer not None):
        """Read the ServerKeyExchange payload."""
        if buffer.size == 0:
            buffer = b"\0"
        cdef const unsigned char* first = &buffer[0]
        cdef const unsigned char* end = &buffer[-1] + 1
        _exc.check_error(_pk.mbedtls_ecdh_read_params(
            &self._ctx, &first, end))


cdef class ECDHNaive(ECDHBase):

    """Naive ECDH key exchange.

    Args:
        (Curve, optional): b'curve25519' or b'curve448'.

    """
    def __init__(self, curve=None):
        if curve is None:
            curve = Curve.CURVE25519
        curve = Curve(curve)
        if curve not in {Curve.CURVE25519, Curve.CURVE448}:
            raise ValueError("ECDHNaive only supports curve25519 and curve448")
        super().__init__(curve)

    def generate(self):
        """Generate a public key.

        Return:
            ECPoint: public key.

        """
        _exc.check_error(_pk.mbedtls_ecdh_gen_public(
            &self._ctx.grp, &self._ctx.d, &self._ctx.Q,
            &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
        return self.public_key

    def import_peer_public(self, pubkey):
        self.import_peers_public(pubkey.x)

    def import_peers_public(self, _pk.ECPoint pubkey):
        """Read peer public key."""
        _exc.check_error(_pk.mbedtls_ecp_copy(&self._ctx.Qp, &pubkey._ctx))

    def generate_secret(self):
        """Generate the shared secret."""
        _exc.check_error(_pk.mbedtls_ecdh_compute_shared(
            &self._ctx.grp, &self._ctx.z, &self._ctx.Qp, &self._ctx.d,
            &_rnd.mbedtls_ctr_drbg_random, &__rng._ctx))
        return _mpi.from_mpi_p(&self._ctx.z)



# SPDX-License-Identifier: MIT
# Copyright (c) 2018, Mathias Laurin

"""A ring buffer.

See Also:
    https://github.com/dhess/c-ringbuf/ Drew Hess' ring
    buffer C implementation was a great help in the
    development of this module.

"""


from libc.stdlib cimport malloc, free
from libc.string cimport memcpy

cimport mbedtls._ringbuf as _rb

import cython


cdef c_init(_rb.ring_buffer_ctx *ctx, size_t maxlen):
    # ringbuf_new()
    ctx._size = maxlen + 1
    ctx.buf = <unsigned char *>malloc(ctx._size * sizeof(unsigned char))
    if not ctx.buf:
        raise MemoryError()
    ctx.head = ctx.tail = ctx.buf


cdef void c_free(_rb.ring_buffer_ctx *ctx) nogil:
    # ringbuf_free()
    free(ctx.buf)
    ctx.head = ctx.tail = NULL


cdef void c_clear(_rb.ring_buffer_ctx *ctx) nogil:
    # ringbuf_reset()
    ctx.head = ctx.tail


cdef size_t c_capacity(_rb.ring_buffer_ctx *ctx) nogil:
    # ringbuf_capacity()
    return ctx._size - 1


cdef unsigned char * c_end(ring_buffer_ctx * ctx) nogil:
    # ringbuf_end()
    return &ctx.buf[ctx._size]


cdef size_t c_len(ring_buffer_ctx *ctx) nogil:
    # ringbuf_bytes_used()
    if ctx.head >= ctx.tail:
        return ctx.head - ctx.tail
    else:
        return c_capacity(ctx) - (ctx.tail - ctx.head - 1)


cdef c_peek(ring_buffer_ctx *ctx, size_t amt):
    cdef unsigned char *dst = <unsigned char *>malloc(
        min(amt, c_len(ctx) * sizeof(unsigned char)))
    if not dst:
        raise MemoryError()
    try:
        nread = c_peekinto(ctx, dst, amt)
        return bytes(dst[:nread])
    finally:
        free(dst)


@cython.boundscheck(False)
cdef size_t c_peekinto(
        ring_buffer_ctx *ctx, unsigned char *dst, size_t amt) nogil:
    cdef size_t size = min(amt, c_len(ctx))
    cdef size_t nread = 0
    cdef unsigned char * index = ctx.tail
    while nread != size:
        if index == c_end(ctx):
            index = ctx.buf

        amt = min(<size_t>(c_end(ctx) - index), size - nread)
        memcpy(&dst[nread], index, amt)
        index += amt
        nread += amt

    return nread


cdef c_read(ring_buffer_ctx *ctx, size_t amt):
    cdef unsigned char *dst = <unsigned char *>malloc(
        min(amt, c_len(ctx)) * sizeof(unsigned char))
    if not dst:
        raise MemoryError()
    try:
        nread = c_readinto(ctx, dst, amt)
        return bytes(dst[:nread])
    finally:
        free(dst)


@cython.boundscheck(False)
cdef size_t c_readinto(
        ring_buffer_ctx *ctx, unsigned char *dst, size_t amt) nogil:
    # ringbuf_memcpy_from()
    cdef size_t size = min(amt, c_len(ctx))
    cdef size_t nread = 0
    while nread != size:
        if ctx.tail == c_end(ctx):
            ctx.tail = ctx.buf

        amt = min(<size_t>(c_end(ctx) - ctx.tail), size - nread)
        memcpy(&dst[nread], ctx.tail, amt)
        ctx.tail += amt
        nread += amt

    return nread


cdef size_t c_consume(ring_buffer_ctx *ctx, size_t amt) nogil:
    cdef size_t size = min(amt, c_len(ctx))
    cdef size_t nconsumed = 0
    while nconsumed != size:
        if ctx.tail == c_end(ctx):
            ctx.tail = ctx.buf

        amt = min(<size_t>(c_end(ctx) - ctx.tail), size - nconsumed)
        ctx.tail += amt
        nconsumed += amt

    return nconsumed


@cython.boundscheck(False)
cdef size_t c_write(
        ring_buffer_ctx *ctx, const unsigned char *src, size_t amt) nogil:
    # ringbuf_memcpy_into()
    cdef size_t size = amt
    # if size > c_capacity(ctx) - c_len(ctx):
    #     raise BufferError("Buffer overflow")

    cdef size_t nwritten = 0
    while nwritten != size:
        if ctx.head == c_end(ctx):
            ctx.head = ctx.buf

        amt = min(<size_t>(c_end(ctx) - ctx.head), size - nwritten)
        memcpy(ctx.head, &src[nwritten], amt)
        ctx.head += amt
        nwritten += amt

    return nwritten


cdef class RingBuffer:
    def __init__(self, size_t maxlen, content=b""):
        self.write(content)

    def __cinit__(self, size_t maxlen, content=b""):
        c_init(&self._ctx, maxlen)

    def __dealloc__(self):
        c_free(&self._ctx)

    def __reduce__(self):
        return type(self), (self.maxlen, self.__bytes__())

    @property
    def maxlen(self):
        return c_capacity(&self._ctx)

    def __eq__(self, other):
        # Call `__bytes__()` directly for Python 2.7.
        try:
            return self.__bytes__() == other.__bytes__()
        except AttributeError:
            return self.__bytes__() == bytes(other)

    def __len__(self):
        return c_len(&self._ctx)

    def __bool__(self):
        return not self.empty()

    def __bytes__(self):
        return self.peek(len(self))

    def clear(self):
        c_clear(&self._ctx)

    def full(self):
        return len(self) == self.maxlen

    def empty(self):
        return len(self) == 0

    def peek(self, size_t amt):
        return c_peek(&self._ctx, amt)

    def read(self, size_t amt):
        return c_read(&self._ctx, amt)

    def consume(self, size_t amt):
        return c_consume(&self._ctx, amt)

    def write(self, const unsigned char[:] src not None, amt=None):
        if src.size == 0:
            return 0
        if amt is None:
            amt = src.size
        else:
            amt = min(src.size, amt)
        if amt > self.maxlen - len(self):
            raise BufferError("Buffer overflow")
        return c_write(&self._ctx, &src[0], amt)



# SPDX-License-Identifier: MIT
# Copyright (c) 2016, Elaborated Networks GmbH
# Copyright (c) 2019, Mathias Laurin

"""Advanced Encryption Standard (AES) cipher established by the U.S.
NIST in 2001.

"""


from mbedtls.exceptions import TLSError

from . import _cipher

__all__ = ["block_size", "key_size", "new"]


block_size = 16
key_size = None


def new(key, mode, iv=None, ad=None):
    """Return a `Cipher` object that can perform AES encryption and
    decryption.

    Advanced Encryption Standard (AES) cipher established by the U.S.
    NIST in 2001.

    Parameters:
        key (bytes or None): The key to encrypt decrypt.  If None,
            encryption and decryption are unavailable.
        mode (int): The mode of operation of the cipher.
        iv (bytes or None): The initialization vector (IV).  The IV is
            required for every mode but ECB and CTR where it is ignored.
            If not set, the IV is initialized to all 0, which should not
            be used for encryption.

    """
    mode = _cipher.Mode(mode)
    if mode in {
        _cipher.Mode.ECB,
        _cipher.Mode.CBC,
        _cipher.Mode.CFB,
        _cipher.Mode.OFB,
        _cipher.Mode.CTR,
        _cipher.Mode.GCM,
        _cipher.Mode.CCM,
    }:
        if len(key) not in {16, 24, 32}:
            raise TLSError(
                msg="key size must 16, 24, or 32 bytes, got %i" % len(key)
            )
    elif mode is _cipher.Mode.XTS:
        if len(key) not in {32, 64}:
            raise TLSError(
                msg="key size must 32, or 64 bytes, got %i" % len(key)
            )
    else:
        raise TLSError(msg="unsupported mode %r" % mode)
    if mode is _cipher.Mode.XTS:
        name = ("AES-%i-%s" % (len(key) * 4, mode.name)).encode("ascii")
    else:
        name = (
            "AES-%i-%s%s"
            % (
                len(key) * 8,
                mode.name,
                "128" if mode is _cipher.Mode.CFB else "",
            )
        ).encode("ascii")
    if mode in {_cipher.Mode.GCM, _cipher.Mode.CCM}:
        return _cipher.AEADCipher(name, key, mode, iv, ad)
    return _cipher.Cipher(name, key, mode, iv)



# SPDX-License-Identifier: MIT
# Copyright (c) 2018, Mathias Laurin

"""Declaration from `mbedtls/bignum.h`."""


cdef extern from "mbedtls/bignum.h" nogil:
    int MBEDTLS_MPI_MAX_SIZE

    # Multi-precision integer library
    # -------------------------------
    ctypedef struct mbedtls_mpi:
        pass

    ctypedef enum mbedtls_mpi_sint:
        pass

    # mbedtls_mpi
    # -----------
    void mbedtls_mpi_init( mbedtls_mpi *X )
    void mbedtls_mpi_free( mbedtls_mpi *X );

    # mbedtls_mpi_grow
    # mbedtls_mpi_shrink

    int mbedtls_mpi_copy(mbedtls_mpi *X, const mbedtls_mpi *Y)

    # mbedtls_mpi_swap
    # mbedtls_mpi_safe_cond_assign
    # mbedtls_mpi_safe_cond_swap
    # mbedtls_mpi_lset  // limited to 64-bits
    # mbedtls_mpi_get_bit
    # mbedtls_mpi_set_bit
    # mbedtls_mpi_lsb

    size_t mbedtls_mpi_bitlen(const mbedtls_mpi *X)
    size_t mbedtls_mpi_size(const mbedtls_mpi *X)

    # mbedtls_mpi_read_string
    # mbedtls_mpi_write_string
    # mbedtls_mpi_read_file
    # mbedtls_mpi_write_file

    int mbedtls_mpi_read_binary(
        mbedtls_mpi *X,
        const unsigned char *buf,
        size_t buflen)
    int mbedtls_mpi_write_binary(
        mbedtls_mpi *X,
        unsigned char *buf,
        size_t buflen)

    int mbedtls_mpi_shift_l(
        mbedtls_mpi *X,
        size_t count)
    int mbedtls_mpi_shift_r(
        mbedtls_mpi *X,
        size_t count)
    # mbedtls_mpi_cmp_abs
    int mbedtls_mpi_cmp_mpi(
        const mbedtls_mpi *X,
        const mbedtls_mpi *Y)
    # mbedtls_mpi_cmp_int
    # mbedtls_mpi_add_abs
    # mbedtls_mpi_sub_abs
    int mbedtls_mpi_add_mpi(
        mbedtls_mpi *X,
        const mbedtls_mpi *A,
        const mbedtls_mpi *B)
    int mbedtls_mpi_sub_mpi(
        mbedtls_mpi *X,
        const mbedtls_mpi *A,
        const mbedtls_mpi *B)
    # mbedtls_mpi_add_int
    # mbedtls_mpi_sub_int
    int mbedtls_mpi_mul_mpi(
        mbedtls_mpi *X,
        const mbedtls_mpi *A,
        const mbedtls_mpi *B)
    # mbedtls_mpi_mul_int
    int mbedtls_mpi_div_mpi(
        mbedtls_mpi *Q,
        mbedtls_mpi *R,
        const mbedtls_mpi *A,
        const mbedtls_mpi *B)
    # mbedtls_mpi_div_int
    int mbedtls_mpi_mod_mpi(
        mbedtls_mpi *X,
        const mbedtls_mpi *A,
        const mbedtls_mpi *B)
    # mbedtls_mpi_mod_int
    int mbedtls_mpi_exp_mod(
        mbedtls_mpi *X,
        const mbedtls_mpi *A,
        const mbedtls_mpi *E,
        const mbedtls_mpi *N,
        mbedtls_mpi *_RR)
    int mbedtls_mpi_fill_random(
        mbedtls_mpi *X, size_t size,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng )
    # mbedtls_mpi_gcd
    # mbedtls_mpi_inv_mod
    int mbedtls_mpi_is_prime(
        mbedtls_mpi *X,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng )
    int mbedtls_mpi_gen_prime(
        mbedtls_mpi *X, size_t size, int dh_flag,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng )


cdef from_mpi_p(mbedtls_mpi *)


cdef class MPI:
    cdef mbedtls_mpi _ctx
    cdef size_t _len(self)



# SPDX-License-Identifier: MIT
# Copyright (c) 2019, Mathias Laurin

"""Run-time version information"""

from libc.stdlib cimport malloc, free
from libc.string cimport strcpy

cimport mbedtls.version as _ver

from collections import namedtuple


mbedtls_version = namedtuple("mbedtls_version", "major minor micro")


cdef __version_info():
    """Returns the version as a tuple."""
    cdef unsigned int version = _ver.mbedtls_version_get_number()
    major = version >> 24 & 0xff
    minor = version >> 16 & 0xff
    micro = version >> 8  & 0xff
    assert version & 0xff == 0
    return mbedtls_version(major, minor, micro)


cdef __version():
    """Return the version as a string."""
    cdef char *output = <char *>malloc(18 * sizeof(char))
    cdef bytes buffer
    if not output:
        raise MemoryError()
    try:
        _ver.mbedtls_version_get_string_full(output)
        buffer = output
        return buffer.decode("ascii")
    finally:
        free(output)


cdef _has_feature(feature):
    feature_ = feature.encode("ascii")
    cdef char *c_feature = feature_
    cdef int result = _ver.mbedtls_version_check_feature(&c_feature[0])
    if result == -2:
        raise ValueError("%s not supported" % feature)
    return result == 0


def has_feature(feature):
    feature = feature.upper()
    if not feature.startswith("MBEDTLS_"):
        feature = "MBEDTLS_" + feature
    result = _has_feature(feature)
    if result is True:
        return result
    elif not feature.endswith("_C"):
        return _has_feature(feature + "_C")
    else:
        return False


version_info = __version_info()
version = __version()



import pickle
from random import randint

import pytest

from mbedtls._ringbuf import RingBuffer


class TestRingBuf:
    @pytest.fixture(params=[2000, 16384])
    def maxlen(self, request):
        return request.param

    @pytest.fixture(params=[10, 100, 1000, 1997, 1998, 1999, 2000])
    def chunk_size(self, request):
        return request.param

    @pytest.fixture
    def buffer(self, maxlen):
        return RingBuffer(maxlen)

    @pytest.fixture
    def randomize_start(self, buffer, maxlen, randbytes):
        # Randomize start of the buffer.
        size = randint(0, maxlen)
        buffer.write(randbytes(size))
        consumed = buffer.consume(size)
        assert buffer.empty()
        assert consumed == size

    def test_initial_conditions(self, buffer, maxlen):
        assert not buffer
        assert len(buffer) == 0
        assert buffer.maxlen == maxlen
        assert buffer.empty()
        assert not buffer.full()

        assert buffer == b""

    @pytest.mark.usefixtures("randomize_start")
    def test_clear(self, buffer, maxlen, randbytes):
        data = randbytes(maxlen)
        written = buffer.write(data)
        buffer.clear()

        assert written == maxlen
        assert not buffer
        assert len(buffer) == 0
        assert buffer.maxlen == maxlen
        assert buffer.empty()
        assert not buffer.full()

        assert buffer == b""

    @pytest.mark.usefixtures("randomize_start")
    def test_write_empty(self, buffer):
        written = buffer.write(b"")
        assert not written
        assert not buffer
        assert len(buffer) == 0
        assert buffer.empty()
        assert not buffer.full()

        assert buffer == b""

    @pytest.mark.usefixtures("randomize_start")
    def test_pickle(self, buffer, maxlen, randbytes):
        data = randbytes(maxlen // 2)
        written = buffer.write(data)

        other = pickle.loads(pickle.dumps(buffer))
        assert other
        assert other == buffer

    @pytest.mark.usefixtures("randomize_start")
    def test_consume_zero(self, buffer, maxlen, randbytes):
        data = randbytes(maxlen)
        written = buffer.write(data)
        buffer.consume(0)

        assert buffer.full()
        assert buffer == data

    @pytest.mark.usefixtures("randomize_start")
    def test_peek_zero(self, buffer, maxlen, randbytes):
        data = randbytes(maxlen)
        written = buffer.write(data)

        peeked = buffer.peek(0)
        assert peeked == b""

        assert buffer.full()
        assert buffer == data

    @pytest.mark.usefixtures("randomize_start")
    def test_read_zero(self, buffer, maxlen, randbytes):
        data = randbytes(maxlen)
        written = buffer.write(data)

        assert written == maxlen
        assert buffer.read(0) == b""

    @pytest.mark.usefixtures("randomize_start")
    def test_write_full(self, buffer, maxlen, randbytes):
        data = randbytes(maxlen)
        written = buffer.write(data)
        assert written == maxlen
        assert buffer
        assert len(buffer) == maxlen
        assert buffer.maxlen == maxlen
        assert not buffer.empty()
        assert buffer.full()

        assert buffer == data

    @pytest.mark.usefixtures("randomize_start")
    def test_write_chunks(self, buffer, maxlen, chunk_size, randbytes):
        data = randbytes(maxlen)
        written = 0
        for index in range(0, maxlen, chunk_size):
            written += buffer.write(data[index : index + chunk_size])

        assert written == maxlen
        assert buffer.full()
        assert buffer == data

    @pytest.mark.usefixtures("randomize_start")
    def test_peek(self, buffer, maxlen, chunk_size, randbytes):
        data = randbytes(maxlen)
        written = buffer.write(data)
        for index in range(0, maxlen, chunk_size):
            assert buffer.peek(index) == data[0:index]

        assert written == maxlen
        assert buffer == data

    @pytest.mark.usefixtures("randomize_start")
    def test_read_chunks(self, buffer, maxlen, chunk_size, randbytes):
        data = randbytes(maxlen)
        written = buffer.write(data)
        for index in range(0, maxlen, chunk_size):
            assert len(buffer) == maxlen - index
            assert buffer.read(chunk_size) == data[index : index + chunk_size]
        assert written == maxlen
        assert buffer.empty()

    @pytest.mark.usefixtures("randomize_start")
    def test_read_write_with_wraparound(
        self, buffer, maxlen, chunk_size, randbytes
    ):
        wraparound = 5
        written = 0
        for _ in range(wraparound * maxlen // chunk_size):
            data = randbytes(chunk_size)
            written += buffer.write(data)
            assert buffer.read(chunk_size) == data
            assert buffer.empty()
        assert written >= maxlen * (wraparound - 1)

    @pytest.mark.usefixtures("randomize_start")
    def test_read_write_with_wraparound_long(self, buffer, maxlen, randbytes):
        size = randint(2, 42) * maxlen
        step = randint(1, maxlen)
        written = 0
        for idx in range(0, size, step):
            data = randbytes(step)
            amt = buffer.write(data)
            written += amt
            assert amt == step
            assert amt == len(data)
            assert written % step == 0
            assert buffer.read(len(data)) == data
            assert buffer.empty()

    @pytest.mark.usefixtures("randomize_start")
    def test_write_overflow_raises_buffererror(
        self, buffer, maxlen, randbytes
    ):
        data = randbytes(maxlen + 1)
        with pytest.raises(BufferError):
            buffer.write(data)
        assert not buffer



import pytest


@pytest.mark.xfail(reason="Test assertion rewriting")
class TestMemoryviewAssertion:
    @pytest.fixture
    def value(self):
        return bytes(_ % 256 for _ in range(10000))

    @pytest.fixture
    def text(self, value):
        return value.decode("latin1")

    @pytest.fixture
    def memview(self, value):
        return memoryview(value)

    def test_memview_and_memview(self, memview):
        assert memview == memview[::-1]

    def test_bytes_and_shorter_memview(self, value, memview):
        assert value == memview[:-1]

    def test_shorter_memview_and_bytes(self, value, memview):
        assert memview[:-1] == value

    def test_bytes_and_longer_memview(self, value, memview):
        assert value[:-1] == memview

    def test_longer_memview_and_bytes(self, value, memview):
        assert memview == value[:-1]

    def test_memview_and_str(self, text, memview):
        assert memview == text

    def test_str_and_memview(self, text, memview):
        assert text == memview



# SPDX-License-Identifier: MIT
# Copyright (c) 2018, Mathias Laurin

"""Declarations from `mbedtls/ssl.h`."""


cimport mbedtls._ringbuf as _rb
cimport mbedtls.pk as _pk
cimport mbedtls.x509 as _x509


cdef:
    enum:
        MBEDTLS_SSL_TRANSPORT_STREAM = 0
        MBEDTLS_SSL_TRANSPORT_DATAGRAM = 1

    enum:
        MBEDTLS_SSL_PRESET_DEFAULT = 0
        MBEDTLS_SSL_PRESET_SUITEB = 2

    enum:
        MBEDTLS_SSL_VERIFY_NONE = 0
        MBEDTLS_SSL_VERIFY_OPTIONAL = 1
        MBEDTLS_SSL_VERIFY_REQUIRED = 2

    enum:
        MBEDTLS_SSL_ANTI_REPLAY_DISABLED = 0
        MBEDTLS_SSL_ANTI_REPLAY_ENABLED = 1

    enum:
        # Message type
        MBEDTLS_SSL_MSG_CHANGE_CIPHER_SPEC = 20  # 0x14
        MBEDTLS_SSL_MSG_ALERT = 21      # 0x15
        MBEDTLS_SSL_MSG_HANDSHAKE = 22  # 0x16
        MBEDTLS_SSL_MSG_APPLICATION_DATA = 23  # 0x17

    enum:
        MBEDTLS_SSL_HELLO_REQUEST
        MBEDTLS_SSL_CLIENT_HELLO
        MBEDTLS_SSL_SERVER_HELLO
        MBEDTLS_SSL_SERVER_CERTIFICATE
        MBEDTLS_SSL_SERVER_KEY_EXCHANGE
        MBEDTLS_SSL_CERTIFICATE_REQUEST
        MBEDTLS_SSL_SERVER_HELLO_DONE
        MBEDTLS_SSL_CLIENT_CERTIFICATE
        MBEDTLS_SSL_CLIENT_KEY_EXCHANGE
        MBEDTLS_SSL_CERTIFICATE_VERIFY
        MBEDTLS_SSL_CLIENT_CHANGE_CIPHER_SPEC
        MBEDTLS_SSL_CLIENT_FINISHED
        MBEDTLS_SSL_SERVER_CHANGE_CIPHER_SPEC
        MBEDTLS_SSL_SERVER_FINISHED
        MBEDTLS_SSL_FLUSH_BUFFERS
        MBEDTLS_SSL_HANDSHAKE_WRAPUP
        MBEDTLS_SSL_HANDSHAKE_OVER
        MBEDTLS_SSL_SERVER_NEW_SESSION_TICKET
        MBEDTLS_SSL_SERVER_HELLO_VERIFY_REQUEST_SENT

    enum:
        MBEDTLS_SSL_IS_CLIENT
        MBEDTLS_SSL_IS_SERVER

    enum:
        MBEDTLS_ERR_NET_CONN_RESET = -0x0050
        MBEDTLS_ERR_SSL_WANT_READ = -0x6900
        MBEDTLS_ERR_SSL_WANT_WRITE = -0x6880
        MBEDTLS_ERR_SSL_CLIENT_RECONNECT = -0x6780
        MBEDTLS_ERR_SSL_PEER_CLOSE_NOTIFY = -0x7880
        MBEDTLS_ERR_SSL_BUFFER_TOO_SMALL = -0x6a00
        MBEDTLS_ERR_SSL_HELLO_VERIFY_REQUIRED = -0x6a80
        MBEDTLS_ERR_SSL_BAD_INPUT_DATA = -0x7100


cdef extern from "mbedtls/debug.h" nogil:
    void mbedtls_debug_set_threshold(int threshold)


cdef extern from "mbedtls/timing.h" nogil:
    # This provides callbacks for DTLS with blocking IO.

    ctypedef struct mbedtls_timing_hr_time:
        pass

    ctypedef struct mbedtls_timing_delay_context:
        mbedtls_timing_hr_time timer
        int int_ms
        int fin_ms

    # extern volatile int mbedtls_timing_alarmed

    # unsigned long mbedtls_timing_hardclock()
    # unsigned long mbedtls_timing_get_timer(
    #     mbedtls_timing_hr_time *,
    #     int reset,
    # )
    # void mbedtls_set_alarm(int seconds)

    # mbedtls_ssl_set_timer_t callback
    void mbedtls_timing_set_delay(void *data, int int_ms, int fin_ms)
    # mbedtls_ssl_get_timer_t callback
    int mbedtls_timing_get_delay(void *data)


cdef extern from "mbedtls/ssl_internal.h" nogil:
    ctypedef struct mbedtls_ssl_transform:
        pass

    ctypedef struct mbedtls_ssl_handshake_params:
        int sig_alg
        int verify_sig_alg
        # Diffie-Hellman key exchange:
        # mbedtls_dhm_context dhm_ctx
        _pk.mbedtls_ecdh_context ecdh_ctx
        # EC-J-Pake (not very much used anymore)
        # mbedtls_ecjpake_context ecjpake_ctx
        mbedtls_ssl_key_cert *key_cert

    ctypedef struct mbedtls_ssl_key_cert:
        _x509.mbedtls_x509_crt *cert
        _pk.mbedtls_pk_context *key
        mbedtls_ssl_key_cert *next


cdef extern from "mbedtls/ssl.h" nogil:
    # Defined here
    # ------------
    # ctypedef enum mbedtls_ssl_states: pass

    ctypedef struct mbedtls_ssl_session:
        pass

    ctypedef struct mbedtls_ssl_config:
        # set_certificate_chain
        mbedtls_ssl_key_cert *key_cert
        # set_ciphers
        const int *ciphersuite_list[4]
        # set_inner_protocols
        const char **alpn_list
        # set_lowest_supported_version/set_highest_supported_version
        unsigned char max_major_ver
        unsigned char max_minor_ver
        unsigned char min_major_ver
        unsigned char min_minor_ver
        # set_anti_replay
        unsigned int anti_replay
        # set_handshake_timeout
        unsigned int hs_timeout_min
        unsigned int hs_timeout_max

        unsigned int endpoint
        unsigned int transport
        # set_validate_certificates
        unsigned int authmode

        # set_trust_store
        _x509.mbedtls_x509_crt *ca_chain
        _x509.mbedtls_x509_crt *ca_crl
        # set_sni_callback
        # f_sni / p_sni
        # for mbedtls_ssl_conf_psk_cb
        void *p_psk
        unsigned char *psk
        size_t psk_len
        unsigned char *psk_identity
        size_t psk_identity_len

    ctypedef struct mbedtls_ssl_context:
        const mbedtls_ssl_config *conf
        int state
        char *hostname
        unsigned char *cli_id
        size_t cli_id_len

    # Callback types
    # --------------
    ctypedef int(*mbedtls_ssl_send_p)(void*, const unsigned char*, size_t)
    ctypedef int(*mbedtls_ssl_recv_p)(void*, unsigned char*, size_t)
    ctypedef int(*mbedtls_ssl_recv_timeout_p)(
        void*, unsigned char*, size_t, int)

    ctypedef void(*mbedtls_ssl_set_timer_t)(void *ctx, int int_ms, int fin_ms)
    ctypedef int(*mbedtls_ssl_get_timer_t)(void *ctx)
    ctypedef int(*mbedtls_ssl_cookie_write_t)(
        void *ctx,
        unsigned char **p, unsigned char *end,
        const unsigned char *info, size_t ilen)
    ctypedef int(*mbedtls_ssl_cookie_check_t)(
        void *ctx,
        const unsigned char *cookie, size_t clen,
        const unsigned char *info, size_t ilen)
    # mbedtls_ssl_ticket_write_t
    # mbedtls_ssl_ticket_parse_t
    # mbedtls_ssl_export_keys_t

    # Free functions
    # --------------
    const int* mbedtls_ssl_list_ciphersuites()
    const char* mbedtls_ssl_get_ciphersuite_name(const int ciphersuite_id)
    int mbedtls_ssl_get_ciphersuite_id(const char *ciphersuite_name)

    # mbedtls_ssl_config
    # ------------------
    void mbedtls_ssl_conf_endpoint(mbedtls_ssl_config *conf, int endpoint)
    # void mbedtls_ssl_conf_transport(mbedtls_ssl_config *conf, int transport)

    void mbedtls_ssl_conf_authmode(mbedtls_ssl_config *conf, int authmode)
    void mbedtls_ssl_conf_ciphersuites(
        mbedtls_ssl_config *conf,
        const int* ciphersuites)

    # DTLS only
    # ---------
    void mbedtls_ssl_conf_dtls_anti_replay(
        mbedtls_ssl_config *conf,
        char mode)
    # mbedtls_ssl_conf_dtls_badmac_limit
    void mbedtls_ssl_conf_handshake_timeout(
        mbedtls_ssl_config *conf,
        int min, int max)
    # mbedtls_ssl_conf_ciphersuites_for_version
    # mbedtls_ssl_conf_cert_profile

    # TLS + DTLS
    # ----------
    void mbedtls_ssl_conf_ca_chain(
        mbedtls_ssl_config *conf,
        _x509.mbedtls_x509_crt *ca_chain,
        _x509.mbedtls_x509_crl *ca_crl)
    int mbedtls_ssl_conf_own_cert(
        mbedtls_ssl_config *conf,
        _x509.mbedtls_x509_crt *own_cert,
        _pk.mbedtls_pk_context *pk_key)

    int mbedtls_ssl_conf_psk(
        mbedtls_ssl_config *conf,
        const unsigned char *psk,
        size_t psk_len,
        const unsigned char *psk_identity,
        size_t psk_identity_len)
    # mbedtls_ssl_conf_dh_param
    # mbedtls_ssl_conf_dh_param_ctx
    # mbedtls_ssl_conf_dhm_min_bitlen
    # mbedtls_ssl_conf_curves
    # mbedtls_ssl_conf_sig_hashes
    int mbedtls_ssl_conf_alpn_protocols(
        mbedtls_ssl_config *conf,
        const char **protos)
    void mbedtls_ssl_config_init(mbedtls_ssl_config *conf)
    int mbedtls_ssl_config_defaults(
        mbedtls_ssl_config *conf,
        int endpoint,
        int transport,
        int preset)
    void mbedtls_ssl_config_free(mbedtls_ssl_config *conf)

    # mbedtls_ssl_config: set callbacks
    # ---------------------------------
    # mbedtls_ssl_conf_verify  // optional

    void mbedtls_ssl_conf_rng(
        mbedtls_ssl_config *conf,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng )
    void mbedtls_ssl_conf_dbg(
        mbedtls_ssl_config *conf,
        void (*f_dbg)(void *, int, const char *, int, const char *),
        void  *p_dbg )

    # mbedtls_ssl_conf_read_timeout
    # mbedtls_ssl_conf_session_tickets_cb
    # mbedtls_ssl_conf_export_keys_cb

    void mbedtls_ssl_conf_dtls_cookies(
        mbedtls_ssl_config *conf,
        mbedtls_ssl_cookie_write_t f_cookie_write,
        mbedtls_ssl_cookie_check_t f_cookie_check,
        void *p_cookie,
    )

    # mbedtls_ssl_conf_session_cache
    void mbedtls_ssl_conf_psk_cb(
        mbedtls_ssl_config *conf,
        int (*f_psk)(void *, mbedtls_ssl_context *, const unsigned char *, size_t),
        void *psk_store)
    void mbedtls_ssl_conf_sni(
        mbedtls_ssl_config *conf,
        int (*f_sni)(void *, mbedtls_ssl_context *, const unsigned char*,
                     size_t),
        void* p_sni)
    void mbedtls_ssl_conf_max_version(
        mbedtls_ssl_config *conf,
        int major,
        int minor)
    void mbedtls_ssl_conf_min_version(
        mbedtls_ssl_config *conf,
        int major,
        int minor)
    # mbedtls_ssl_conf_fallback
    # mbedtls_ssl_conf_encrypt_then_mac
    # mbedtls_ssl_conf_extended_master_secret
    # mbedtls_ssl_conf_arc4_support
    # mbedtls_ssl_conf_max_frag_len
    # mbedtls_ssl_conf_truncated_hmac
    # mbedtls_ssl_conf_cbc_record_splitting
    # mbedtls_ssl_conf_session_tickets
    void mbedtls_ssl_conf_renegotiation(
        mbedtls_ssl_config *conf,
        int renegotiation)
    # mbedtls_ssl_conf_legacy_renegotiation
    # mbedtls_ssl_conf_renegotiation_enforced
    # mbedtls_ssl_conf_renegotiation_period

    # mbedtls_ssl_context
    # -------------------
    void mbedtls_ssl_init(mbedtls_ssl_context *ctx)
    int mbedtls_ssl_setup(
        mbedtls_ssl_context *ctx,
        const mbedtls_ssl_config *conf)
    int mbedtls_ssl_session_reset(mbedtls_ssl_context *ctx)
    void mbedtls_ssl_set_bio(
        mbedtls_ssl_context *ssl,
        void *p_bio,
        mbedtls_ssl_send_p f_send,
        mbedtls_ssl_recv_p f_recv,
        mbedtls_ssl_recv_timeout_p f_recv_timeout)

    void mbedtls_ssl_set_timer_cb(
        # DTLS
        mbedtls_ssl_context *ssl,
        void *p_timer,
        mbedtls_ssl_set_timer_t f_set_timer,
        mbedtls_ssl_get_timer_t f_get_timer)
    int mbedtls_ssl_set_client_transport_id(
        # DTLS
        mbedtls_ssl_context *ssl,
        const unsigned char *info,
        size_t ilen)
    int mbedtls_ssl_set_session(
        const mbedtls_ssl_context *ssl,
        mbedtls_ssl_session *session)
    int mbedtls_ssl_set_hs_psk(
        mbedtls_ssl_context *ssl,
        const unsigned char *psk,
        size_t psk_len)
    int mbedtls_ssl_set_hostname(
        mbedtls_ssl_context *ssl,
        const char *hostname)
    # mbedtls_ssl_set_hs_ecjpake_password
    # mbedtls_ssl_set_hs_own_cert
    # mbedtls_ssl_set_hs_ca_chain
    # mbedtls_ssl_set_hs_authmode
    const char* mbedtls_ssl_get_alpn_protocol(const mbedtls_ssl_context *ctx)
    size_t mbedtls_ssl_get_bytes_avail(const mbedtls_ssl_context *ctx)
    int mbedtls_ssl_get_verify_result(const mbedtls_ssl_context *ssl)
    const char* mbedtls_ssl_get_ciphersuite(const mbedtls_ssl_context *ssl)
    const char* mbedtls_ssl_get_version(const mbedtls_ssl_context *ssl)
    # mbedtls_ssl_get_record_expansion
    size_t mbedtls_ssl_get_max_frag_len(const mbedtls_ssl_context *ssl)
    const _x509.mbedtls_x509_crt *mbedtls_ssl_get_peer_cert(
        const mbedtls_ssl_context *ctx)
    int mbedtls_ssl_get_session(
        const mbedtls_ssl_context *ssl,
        mbedtls_ssl_session *session)
    int mbedtls_ssl_handshake(mbedtls_ssl_context *ctx)
    int mbedtls_ssl_handshake_step(mbedtls_ssl_context *ssl)
    int mbedtls_ssl_renegotiate(mbedtls_ssl_context *ssl)
    int mbedtls_ssl_read(
        mbedtls_ssl_context *ctx,
        unsigned char *buf,
        size_t len)
    int mbedtls_ssl_write(
        mbedtls_ssl_context *ctx,
        const unsigned char *buf,
        size_t len)
    # mbedtls_ssl_send_alert_message
    int mbedtls_ssl_close_notify(mbedtls_ssl_context *ssl)
    void mbedtls_ssl_free(mbedtls_ssl_context *ctx)

    # mbedtls_ssl_session
    # -------------------
    void mbedtls_ssl_session_init(mbedtls_ssl_session *session)
    void mbedtls_ssl_session_free(mbedtls_ssl_session *session)


cdef extern from "mbedtls/ssl_cookie.h" nogil:
    # This provides callbacks for DTLS.

    ctypedef struct mbedtls_ssl_cookie_ctx:
        # mbedtls_md_context_t hmac_ctx
        unsigned long timeout
        # mbedtls_threading_mutex_t mutex

    void mbedtls_ssl_cookie_init(mbedtls_ssl_cookie_ctx *ctx)
    int mbedtls_ssl_cookie_setup(
        mbedtls_ssl_cookie_ctx *ctx,
        int (*f_rng)(void *, unsigned char *, size_t), void *p_rng,
    )
    void mbedtls_ssl_cookie_set_timeout(
        mbedtls_ssl_cookie_ctx *ctx,
        unsigned long delay,
    )
    void mbedtls_ssl_cookie_free(mbedtls_ssl_cookie_ctx *ctx)
    mbedtls_ssl_cookie_write_t mbedtls_ssl_cookie_write
    mbedtls_ssl_cookie_check_t mbedtls_ssl_cookie_check


cdef class _DTLSCookie:
    cdef mbedtls_ssl_cookie_ctx _ctx


cdef class _PSKSToreProxy:
    cdef object _mapping


cdef class _BaseConfiguration:
    cdef mbedtls_ssl_config _ctx
    cdef _chain
    cdef int *_ciphers
    cdef const char **_protos
    cdef _PSKSToreProxy _store
    # cdef'd because we aim at a non-writable structure.
    cdef _set_validate_certificates(self, validate)
    cdef _set_certificate_chain(self, chain)
    cdef _set_ciphers(self, ciphers)
    cdef _set_inner_protocols(self, protocols)
    cdef _set_lowest_supported_version(self, version)
    cdef _set_highest_supported_version(self, version)
    cdef _set_trust_store(self, object store)
    cdef _set_sni_callback(self, callback)
    cdef _set_pre_shared_key(self, psk)
    cdef _set_pre_shared_key_store(self, psk_store)


cdef class TLSConfiguration(_BaseConfiguration):
    pass


cdef class DTLSConfiguration(_BaseConfiguration):
    cdef _DTLSCookie _cookie
    cdef _set_anti_replay(self, mode)
    cdef _set_handshake_timeout(self, minimum, maximum)
    cdef _set_cookie(self, _DTLSCookie cookie)


cdef class TLSSession:
    cdef mbedtls_ssl_session _ctx


cdef struct _C_Buffers:
    _rb.ring_buffer_ctx *out_ctx
    _rb.ring_buffer_ctx *in_ctx


cdef class _BaseContext:
    cdef _BaseConfiguration _conf


cdef class MbedTLSBuffer:
    cdef _BaseContext _context
    cdef mbedtls_ssl_context _ctx
    cdef _rb.RingBuffer _output_buffer
    cdef _rb.RingBuffer _input_buffer
    cdef _C_Buffers _c_buffers
    # DTLS only:
    cdef mbedtls_timing_delay_context _timer



# SPDX-License-Identifier: MIT
# Copyright (c) 2019, Mathias Laurin

"""Declarations from `mbedtls/version.h`."""


cdef extern from "mbedtls/version.h" nogil:
    unsigned int mbedtls_version_get_number()
    # void mbedtls_version_get_string(char *string)
    void mbedtls_version_get_string_full(char *string)
    int mbedtls_version_check_feature(const char *feature)



# SPDX-License-Identifier: MIT
# Copyright (c) 2018, Mathias Laurin

"""Declarations from `mbedtls/x509*.h`.

 - CSR: Certificate signing request parsing and writing.
 - CRL: Certificate revocation list parsing.
 - CRT: Certificate parsing and writing.

"""


cimport mbedtls._md as _md
cimport mbedtls.mpi as _mpi
cimport mbedtls.pk as _pk


cdef extern from "mbedtls/asn1.h" nogil:
    cdef struct mbedtls_asn1_buf:
        int tag
        size_t len
        unsigned char *p
    cdef struct mbedtls_asn1_sequence:
        mbedtls_asn1_buf buf
        mbedtls_asn1_sequence *next
    cdef struct mbedtls_asn1_named_data:
        mbedtls_asn1_buf oid
        mbedtls_asn1_buf val
        mbedtls_asn1_named_data *next
        unsigned char next_merged


cdef extern from "mbedtls/x509.h" nogil:
    ctypedef mbedtls_asn1_buf mbedtls_x509_buf
    ctypedef mbedtls_asn1_named_data mbedtls_x509_name
    ctypedef mbedtls_asn1_sequence mbedtls_x509_sequence
    int mbedtls_x509_dn_gets(
        char *buf, size_t size, const mbedtls_x509_name *dn)
    cdef struct mbedtls_x509_time:
        int year, mon, day
        int hour, min, sec


cdef extern from "mbedtls/x509_crt.h" nogil:
    cdef struct mbedtls_x509_crt:
        mbedtls_x509_buf raw
        mbedtls_x509_buf tbs
        int version
        mbedtls_x509_buf serial
        mbedtls_x509_buf sig_oid
        # mbedtls_x509_buf issuer_raw
        # mbedtls_x509_buf subject_raw
        mbedtls_x509_name issuer
        mbedtls_x509_name subject
        mbedtls_x509_time valid_from
        mbedtls_x509_time valid_to
        _pk.mbedtls_pk_context pk  # public key
        # mbedtls_x509_buf issuer_id
        # mbedtls_x509_buf subject_id
        # mbedtls_x509_buf v3_ext
        mbedtls_x509_sequence subject_alt_names
        # int ext_types
        int ca_istrue  # 1 if this certificate belongs to a CA, 0 otherwise
        int max_pathlen
        unsigned int key_usage
        # mbedtls_x509_sequence ext_key_usage
        # unsigned char ns_cert_type

        mbedtls_x509_buf sig
        _md.mbedtls_md_type_t sig_md
        # mbedtls_pk_type_t sig_pk
        # void *sig_opts
        mbedtls_x509_crt *next

    ctypedef struct mbedtls_x509_crt_profile:
        pass

    ctypedef struct mbedtls_x509write_cert:
        pass

    # mbedtls_x509_crt
    # ----------------
    int mbedtls_x509_crt_parse_der(
        mbedtls_x509_crt *chain,
        const unsigned char *buf,
        size_t buflen)
    int mbedtls_x509_crt_parse(
        mbedtls_x509_crt *chain,
        const unsigned char *buf,
        size_t buflen)
    int mbedtls_x509_crt_parse_file(
        mbedtls_x509_crt *chain,
        const char *path)

    # mbedtls_x509_crt_parse_path

    int mbedtls_x509_crt_info(
        char *buf,
        size_t size,
        const char* prefix,
        const mbedtls_x509_crt *crt)

    # mbedtls_x509_crt_verify_info
    # mbedtls_x509_crt_verify
    # mbedtls_x509_crt_verify_with_profile
    # mbedtls_x509_crt_check_key_usage
    # mbedtls_x509_crt_check_extended_key_usage

    int mbedtls_x509_crt_is_revoked(
        const mbedtls_x509_crt *crt,
        const mbedtls_x509_crl *crl)

    void mbedtls_x509_crt_init(mbedtls_x509_crt *crt)
    void mbedtls_x509_crt_free(mbedtls_x509_crt *crt)

    # mbedtls_x509write_cert
    # ----------------------

    void mbedtls_x509write_crt_init(mbedtls_x509write_cert *ctx)
    void mbedtls_x509write_crt_set_version(
        mbedtls_x509write_cert *ctx,
        int version)
    int mbedtls_x509write_crt_set_serial(
        mbedtls_x509write_cert *ctx,
        _mpi.mbedtls_mpi *serial)

    int mbedtls_x509write_crt_set_validity(
        mbedtls_x509write_cert *ctx,
        const char *not_before,
        const char *not_after)
    int mbedtls_x509write_crt_set_issuer_name(
        mbedtls_x509write_cert *ctx,
        const char *issuer_name)
    int mbedtls_x509write_crt_set_subject_name(
        mbedtls_x509write_cert *ctx,
        const char *subject_name)
    void mbedtls_x509write_crt_set_subject_key(
        mbedtls_x509write_cert *ctx,
        _pk.mbedtls_pk_context *key)
    void mbedtls_x509write_crt_set_issuer_key(
        mbedtls_x509write_cert *ctx,
        _pk.mbedtls_pk_context *key)
    void mbedtls_x509write_crt_set_md_alg(
        mbedtls_x509write_cert *ctx,
        _md.mbedtls_md_type_t md_alg)

    # mbedtls_x509write_crt_set_extension
    int mbedtls_x509write_crt_set_basic_constraints(
        mbedtls_x509write_cert *ctx,
        int is_ca, int max_pathlen)

    int mbedtls_x509write_crt_set_subject_key_identifier(
        mbedtls_x509write_cert *ctx)
    int mbedtls_x509write_crt_set_authority_key_identifier(
        mbedtls_x509write_cert *ctx)

    # mbedtls_x509write_crt_set_key_usage
    # mbedtls_x509write_crt_set_ns_cert_type

    void mbedtls_x509write_crt_free(mbedtls_x509write_cert *ctx)

    int mbedtls_x509write_crt_der(
        mbedtls_x509write_cert *ctx,
        unsigned char *buf, size_t size,
        int (*f_rng)(void *, unsigned char *, size_t),
        void *p_rng)
    int mbedtls_x509write_crt_pem(
        mbedtls_x509write_cert *ctx,
        unsigned char *buf, size_t size,
        int (*f_rng)(void *, unsigned char *, size_t),
        void *p_rng)


cdef extern from "mbedtls/x509_csr.h" nogil:
    # Certificate signing request parsing and writing
    # -----------------------------------------------
    cdef struct mbedtls_x509_csr:
        mbedtls_x509_buf raw
        # mbedtls_x509_buf cri
        int version
        # mbedtls_x509_buf subject_raw
        mbedtls_x509_name subject
        _pk.mbedtls_pk_context pk
        mbedtls_x509_buf sig_oid
        mbedtls_x509_buf sig
        _md.mbedtls_md_type_t sig_md
        # mbedtls_pk_type_t sig_pk
        # void *sig_opts

    ctypedef struct mbedtls_x509write_csr:
        pass

    # mbedtls_x509_csr
    # ----------------
    int mbedtls_x509_csr_parse_der(
        mbedtls_x509_csr *csr,
        const unsigned char *buf,
        size_t buflen)
    int mbedtls_x509_csr_parse(
        mbedtls_x509_csr *csr,
        const unsigned char *buf,
        size_t buflen)
    int mbedtls_x509_csr_parse_file(
        mbedtls_x509_csr *csr,
        const char *path)
    int mbedtls_x509_csr_info(
        char *buf,
        size_t size,
        const char* prefix,
        const mbedtls_x509_csr *csr)
    void mbedtls_x509_csr_init(mbedtls_x509_csr *csr)
    void mbedtls_x509_csr_free(mbedtls_x509_csr *csr)

    # mbedtls_x509write_csr
    # ---------------------
    void mbedtls_x509write_csr_init(mbedtls_x509write_csr *ctx)

    int mbedtls_x509write_csr_set_subject_name(
        mbedtls_x509write_csr *ctx,
        const char *subject_name)
    void mbedtls_x509write_csr_set_key(
        mbedtls_x509write_csr *ctx,
        _pk.mbedtls_pk_context *key)
    void mbedtls_x509write_csr_set_md_alg(
        mbedtls_x509write_csr *ctx,
        _md.mbedtls_md_type_t md_alg)

    # mbedtls_x509write_csr_set_key_usage
    # mbedtls_x509write_csr_set_ns_cert_type
    # mbedtls_x509write_csr_set_extension

    void mbedtls_x509write_csr_free(mbedtls_x509write_csr *ctx)

    int mbedtls_x509write_csr_der(
        mbedtls_x509write_csr *ctx,
        unsigned char *buf, size_t size,
        int (*f_rng)(void *, unsigned char *, size_t),
        void *p_rng)
    int mbedtls_x509write_csr_pem(
        mbedtls_x509write_csr *ctx,
        unsigned char *buf, size_t size,
        int (*f_rng)(void *, unsigned char *, size_t),
        void *p_rng)


cdef extern from "mbedtls/x509_crl.h" nogil:
    # Certificate revocation list parsing
    # -----------------------------------
    ctypedef struct mbedtls_x509_crl_entry:
        mbedtls_x509_buf raw
        mbedtls_x509_buf serial
        mbedtls_x509_time revocation_date
        mbedtls_x509_buf entry_ext
        mbedtls_x509_crl_entry *next

    cdef struct mbedtls_x509_crl:
        mbedtls_x509_buf raw
        mbedtls_x509_buf tbs
        int version
        mbedtls_x509_buf sig_oid
        mbedtls_x509_buf issuer_raw
        mbedtls_x509_name issuer
        mbedtls_x509_time this_update
        mbedtls_x509_time next_update
        mbedtls_x509_crl_entry entry
        mbedtls_x509_buf crl_ext
        mbedtls_x509_buf sig_oid2
        mbedtls_x509_buf sig
        _md.mbedtls_md_type_t sig_md
        # mbedtls_pk_type_t sig_pk
        # void *sig_opts
        mbedtls_x509_crl *next

    # mbedtls_x509_crl
    # ----------------
    int mbedtls_x509_crl_parse_der(
        mbedtls_x509_crl *chain,
        const unsigned char *buf,
        size_t buflen)
    int mbedtls_x509_crl_parse(
        mbedtls_x509_crl *chain,
        const unsigned char *buf,
        size_t buflen)
    int mbedtls_x509_crl_parse_file(
        mbedtls_x509_crl *chain,
        const char *path)
    int mbedtls_x509_crl_info(
        char *buf,
        size_t size,
        const char *prefix,
        const mbedtls_x509_crl *crl)
    void mbedtls_x509_crl_init(mbedtls_x509_crl *crl)
    void mbedtls_x509_crl_free(mbedtls_x509_crl *crl)


cdef class Certificate:
    pass


cdef class CRT(Certificate):
    cdef mbedtls_x509_crt _ctx
    cdef set_next(self, CRT crt)
    cdef unset_next(self)
    cdef CRT _next


cdef class _CRTWriter:
    cdef mbedtls_x509write_cert _ctx


cdef class CSR(Certificate):
    cdef mbedtls_x509_csr _ctx


cdef class _CSRWriter:
    cdef mbedtls_x509write_csr _ctx


cdef class CRL(Certificate):
    cdef mbedtls_x509_crl _ctx
    cdef set_next(self, CRL crl)
    cdef unset_next(self)
    cdef CRL _next



# SPDX-License-Identifier: MIT
# Copyright (c) 2016, Elaborated Networks GmbH
# Copyright (c) 2019, Mathias Laurin

"""Random number generator (RNG) wrapper."""


from libc.stdlib cimport malloc, free

cimport mbedtls._platform as _plt
cimport mbedtls._random as _rnd

import numbers as _numbers

import mbedtls.mpi as _mpi
from mbedtls.exceptions import check_error


BPF = 53  # Number of bits in a float
RECIP_BPF = 2**-BPF


cdef class _Entropy:
    def __cinit__(self):
        """Initialize the context."""
        _rnd.mbedtls_entropy_init(&self._ctx)

    def __dealloc__(self):
        """Free and clear the context."""
        _rnd.mbedtls_entropy_free(&self._ctx)

    def __getstate__(self):
        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")

    def gather(self):
        """Trigger an extra gather poll for the accumulator."""
        _rnd.mbedtls_entropy_gather(&self._ctx)

    def retrieve(self, size_t length):
        """Retrieve entropy from the accumulator."""
        cdef unsigned char *output = <unsigned char *> malloc(
            length * sizeof(unsigned char)
        )
        if not output:
            raise MemoryError()
        try:
            check_error(_rnd.mbedtls_entropy_func(&self._ctx, output, length))
            return output[:length]
        finally:
            free(output)

    def update(self, const unsigned char[:] data):
        """Add data to the accumulator manually."""
        check_error(
            _rnd.mbedtls_entropy_update_manual(
                &self._ctx, &data[0], data.shape[0]
            )
        )


cdef class Random:
    def __init__(self):
        self._entropy = _Entropy()
        check_error(
            _rnd.mbedtls_ctr_drbg_seed(
                &self._ctx,
                &_rnd.mbedtls_entropy_func,
                &self._entropy._ctx,
                NULL, 0
            )
        )

    def __cinit__(self):
        """Initialize the context."""
        _rnd.mbedtls_ctr_drbg_init(&self._ctx)

    def __dealloc__(self):
        """Free and clear the context."""
        _rnd.mbedtls_ctr_drbg_free(&self._ctx)

    def __getstate__(self):
        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")

    @property
    def _entropy(self):
        return self._entropy

    def _reseed(self, const unsigned char[:] data=None):
        """Reseed the RNG."""
        if data is None:
            check_error(_rnd.mbedtls_ctr_drbg_reseed(&self._ctx, NULL, 0))
        else:
            check_error(
                _rnd.mbedtls_ctr_drbg_reseed(&self._ctx, &data[0], data.size)
            )

    def urandom(self, size_t length):
        """Returns `length` random bytes."""
        cdef unsigned char *output = <unsigned char *> malloc(
            length * sizeof(unsigned char)
        )
        if not output:
            raise MemoryError()
        try:
            check_error(
                _rnd.mbedtls_ctr_drbg_random(&self._ctx, output, length)
            )
            ret = output[:length]
            _plt.mbedtls_platform_zeroize(output, length)
            return ret
        finally:
            free(output)

    def randbelow(self, upper_bound):
        """Return a random int in the range [0, n).

        Raises ValueError if n <= 0.

        """
        if upper_bound <= 0:
            raise ValueError("Upper bound must be positive.")
        kk = upper_bound.bit_length()
        rr = self.getrandbits(kk)
        while rr >= upper_bound:
            rr = self.getrandbits(kk)
        return rr

    def random(self):
        """Return the next random floating point number."""
        # Algorithm taken from Python's secrets and random libraries.
        return float(
            _mpi.MPI.from_bytes(self.urandom(7), "big") >> 3
        ) * RECIP_BPF

    def getrandbits(self, k):
        """Generate an int with `k` random bits."""
        # Algorithm adapted from Python's secrets and random libraries.
        if k <= 0:
            raise ValueError("number of bits must be greater than zero")
        if not isinstance(k, _numbers.Integral):
            raise TypeError("number of bits should be an integer")
        numbytes = (k + 7) // 8
        value = _mpi.MPI.from_bytes(self.urandom(numbytes), "big")
        # Trim excess bits:
        extra_bits = value.bit_length() - k
        return value >> (0 if extra_bits <= 0 else extra_bits)

    def choice(self, seq):
        """Return a random element from `seq`."""
        try:
            ii = self.randbelow(len(seq))
        except ValueError:
            raise IndexError("Cannot choose from an empty sequence")
        return seq[ii]


cdef Random __rng = Random()


cpdef Random default_rng():
    return __rng
